# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.identity import rbac_base

CONF = config.CONF


class IdentityTrustV3RbacTest(rbac_base.BaseIdentityV3RbacTest):

    credentials = ['primary', 'alt']

    @classmethod
    def skip_checks(cls):
        super(IdentityTrustV3RbacTest, cls).skip_checks()
        if not CONF.identity_feature_enabled.trust:
            raise cls.skipException(
                "%s skipped as trust feature isn't enabled" % cls.__name__)

    @classmethod
    def resource_setup(cls):
        super(IdentityTrustV3RbacTest, cls).resource_setup()
        # Use the primary user's credentials for the "trustor_user_id", since
        # user_id:%(trust.trustor_user_id)s will thereby evaluate to
        # "primary user's user_id:primary user's user_id" which evaluates to
        # true.
        cls.trustor_user_id = cls.os_primary.credentials.user_id
        cls.trustor_project_id = cls.os_primary.credentials.project_id
        cls.trustee_user_id = cls.setup_test_user()['id']

        # The "unauthorized_user_id" does not have permissions to create a
        # trust because the user_id in "user_id:%(trust.trustor_user_id)s" (the
        # policy rule for creating a trust) corresponds to the primary user_id
        # not the alt user_id.
        cls.unauthorized_user_id = cls.os_alt.credentials.user_id

        # A role is guaranteed to exist (namely the admin role), because
        # "trustor_user_id" and "trustor_project_id" are the primary tempest
        # user and project, respectively.
        cls.delegated_role_id = cls.roles_client.list_user_roles_on_project(
            cls.trustor_project_id, cls.trustor_user_id)['roles'][0]['id']

        cls.trust = cls.setup_test_trust(trustor_user_id=cls.trustor_user_id,
                                         trustee_user_id=cls.trustee_user_id,
                                         project_id=cls.trustor_project_id,
                                         roles=[{'id': cls.delegated_role_id}])

    @decorators.idempotent_id('7ab595a7-9b71-45fe-91d8-2793b0292f72')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:create_trust"],
        extra_target_data={
            "trust.trustor_user_id": "os_primary.credentials.user_id"
        })
    def test_create_trust(self):
        with self.override_role():
            self.setup_test_trust(trustor_user_id=self.trustor_user_id,
                                  trustee_user_id=self.trustee_user_id)

    @decorators.idempotent_id('bd72d22a-6e11-4840-bd93-17b382e7f0e0')
    @decorators.attr(type=['negative'])
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:create_trust"],
        extra_target_data={
            "trust.trustor_user_id": "os_alt.credentials.user_id"
        })
    def test_create_trust_negative(self):
        # Explicit negative test for identity:create_trust policy action.
        # Assert expected exception is Forbidden and then reraise it.
        with self.override_role():
            e = self.assertRaises(lib_exc.Forbidden, self.setup_test_trust,
                                  trustor_user_id=self.unauthorized_user_id,
                                  trustee_user_id=self.trustee_user_id)
            raise e

    @decorators.idempotent_id('d9a6fd06-08f6-462c-a86c-ce009adf1230')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:delete_trust"])
    def test_delete_trust(self):
        trust = self.setup_test_trust(trustor_user_id=self.trustor_user_id,
                                      trustee_user_id=self.trustee_user_id)

        with self.override_role():
            self.trusts_client.delete_trust(trust['id'])

    @decorators.idempotent_id('f2e32896-bf66-4f4e-89cf-e7fba0ef1f38')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:list_trusts"])
    def test_list_trusts(self):
        with self.override_role():
            self.trusts_client.list_trusts(
                trustor_user_id=self.trustor_user_id)

    @decorators.idempotent_id('3c9ff92f-a73e-4f9b-8865-e017f38c70f5')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:list_roles_for_trust"])
    def test_list_roles_for_trust(self):
        with self.override_role():
            self.trusts_client.list_trust_roles(self.trust['id'])

    @decorators.idempotent_id('3bb4f97b-cecd-4c7d-ad10-b88ee6c5d573')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:get_role_for_trust"])
    def test_show_trust_role(self):
        with self.override_role():
            self.trusts_client.show_trust_role(
                self.trust['id'], self.delegated_role_id)

    @decorators.idempotent_id('0184e0fb-641e-4b52-ab73-81c1ce6ca5c1')
    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:get_trust"])
    def test_show_trust(self):
        with self.override_role():
            self.trusts_client.show_trust(self.trust['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import sys
import time

from oslo_log import log as logging
from oslo_utils import excutils

from tempest import clients
from tempest.common import credentials_factory as credentials
from tempest import config
from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin import rbac_exceptions

CONF = config.CONF
LOG = logging.getLogger(__name__)


class _ValidateListContext(object):
    """Context class responsible for validation of the list functions.

    This class is used in ``override_role_and_validate_list`` function and
    the result of a list function must be assigned to the ``ctx.resources``
    variable.

    Example::

        with self.override_role_and_validate_list(...) as ctx:
            ctx.resources = list_function()

    """
    def __init__(self, admin_resources=None, admin_resource_id=None):
        """Constructor for ``ValidateListContext``.

        Either ``admin_resources`` or ``admin_resource_id`` should be used,
            not both.

        :param list admin_resources: The list of resources received before
            calling the ``override_role_and_validate_list`` function. To
            validate will be used the ``_validate_len`` function.
        :param UUID admin_resource_id: An ID of a resource created before
            calling the ``override_role_and_validate_list`` function. To
            validate will be used the ``_validate_resource`` function.
        :raises RbacValidateListException: if both ``admin_resources`` and
            ``admin_resource_id`` are set or unset.
        """
        self.resources = None
        if admin_resources is not None and not admin_resource_id:
            self._admin_len = len(admin_resources)
            if not self._admin_len:
                raise rbac_exceptions.RbacValidateListException(
                    reason="the list of admin resources cannot be empty")
            self._validate_func = self._validate_len
        elif admin_resource_id and admin_resources is None:
            self._admin_resource_id = admin_resource_id
            self._validate_func = self._validate_resource
        else:
            raise rbac_exceptions.RbacValidateListException(
                reason="admin_resources and admin_resource_id are mutually "
                       "exclusive")

    def _validate_len(self):
        """Validates that the number of resources is less than admin resources.
        """
        if not len(self.resources):
            raise rbac_exceptions.RbacEmptyResponseBody()
        elif self._admin_len > len(self.resources):
            raise rbac_exceptions.RbacPartialResponseBody(body=self.resources)

    def _validate_resource(self):
        """Validates that the admin resource is present in the resources.
        """
        for resource in self.resources:
            if resource['id'] == self._admin_resource_id:
                return
        raise rbac_exceptions.RbacPartialResponseBody(body=self.resources)

    def _validate(self):
        """Calls the proper validation function.

        :raises RbacValidateListException: if the ``ctx.resources`` variable is
            not assigned.
        """
        if self.resources is None:
            raise rbac_exceptions.RbacValidateListException(
                reason="ctx.resources is not assigned")
        self._validate_func()


class RbacUtilsMixin(object):
    """Utility mixin responsible for switching ``os_primary`` role.

    Should be used as a mixin class alongside an instance of
    :py:class:`tempest.test.BaseTestCase` to perform Patrole class setup for a
    base RBAC class. Child classes should not use this mixin.

    Example::

        class BaseRbacTest(rbac_utils.RbacUtilsMixin, base.BaseV2ComputeTest):

            @classmethod
            def setup_clients(cls):
                super(BaseRbacTest, cls).setup_clients()

                cls.hosts_client = cls.os_primary.hosts_client
                ...

    This class is responsible for overriding the value of the primary Tempest
    credential's role (i.e. ``os_primary`` role). By doing so, it is possible
    to seamlessly swap between admin credentials, needed for setup and clean
    up, and primary credentials, needed to perform the API call which does
    policy enforcement. The primary credentials always cycle between roles
    defined by ``CONF.identity.admin_role`` and
    ``CONF.patrole.rbac_test_roles``.
    """

    def __init__(self, *args, **kwargs):
        super(RbacUtilsMixin, self).__init__(*args, **kwargs)

        # Shows if override_role was called.
        self.__override_role_called = False
        # Shows if exception raised during override_role.
        self.__override_role_caught_exc = False

    _admin_role_id = None
    _rbac_role_ids = None
    _project_id = None
    _user_id = None
    _role_map = None
    _role_inferences_mapping = None

    admin_roles_client = None

    @classmethod
    def setup_clients(cls):
        # Intialize the admin roles_client to perform role switching.
        admin_mgr = clients.Manager(
            credentials.get_configured_admin_credentials())
        if CONF.identity_feature_enabled.api_v3:
            admin_roles_client = admin_mgr.roles_v3_client
        else:
            raise lib_exc.InvalidConfiguration(
                "Patrole role overriding only supports v3 identity API.")

        cls.admin_roles_client = admin_roles_client

        cls._project_id = cls.os_primary.credentials.tenant_id
        cls._user_id = cls.os_primary.credentials.user_id
        cls._role_inferences_mapping = cls._prepare_role_inferences_mapping()

        cls._init_roles()

        # Change default role to admin
        cls._override_role(False)
        super(RbacUtilsMixin, cls).setup_clients()

    @classmethod
    def _prepare_role_inferences_mapping(cls):
        """Preparing roles mapping to support role inferences

        Making query to `list-all-role-inference-rules`_ keystone API
        returns all inference rules, which makes it possible to prepare
        roles mapping.

        It walks recursively through the raw data::

            {"role_inferences": [
                {
                  "implies": [{"id": "3", "name": "reader"}],
                  "prior_role": {"id": "2", "name": "member"}
                },
                {
                  "implies": [{"id": "2", "name": "member"}],
                  "prior_role": {"id": "1", "name": "admin"}
                }
              ]
            }

        and converts it to the mapping::

            {
              "2": ["3"],      # "member": ["reader"],
              "1": ["2", "3"]  # "admin": ["member", "reader"]
            }

        .. _list-all-role-inference-rules: https://developer.openstack.org/api-ref/identity/v3/#list-all-role-inference-rules
        """  # noqa: E501
        def process_roles(role_id, data):
            roles = data.get(role_id, set())
            for rid in roles.copy():
                roles.update(process_roles(rid, data))

            return roles

        def convert_data(data):
            res = {}
            for rule in data:
                prior_role = rule['prior_role']['id']
                implies = {r['id'] for r in rule['implies']}
                res[prior_role] = implies
            return res

        raw_data = cls.admin_roles_client.list_all_role_inference_rules()
        data = convert_data(raw_data['role_inferences'])
        res = {}
        for role_id in data:
            res[role_id] = process_roles(role_id, data)
        return res

    def get_all_needed_roles(self, roles):
        """Extending given roles with roles from mapping

        Examples::
            ["admin"] >> ["admin", "member", "reader"]
            ["member"] >> ["member", "reader"]
            ["reader"] >> ["reader"]
            ["custom_role"] >> ["custom_role"]

        :param roles: list of roles
        :return: extended list of roles
        """
        res = set(r for r in roles)
        for role in res.copy():
            role_id = self.__class__._role_map.get(role)
            implied_roles = self.__class__._role_inferences_mapping.get(
                role_id, set())
            role_names = {self.__class__._role_map[rid]
                          for rid in implied_roles}
            res.update(role_names)
        LOG.debug('All needed roles: %s; Base roles: %s', res, roles)
        return list(res)

    @contextlib.contextmanager
    def override_role(self):
        """Override the role used by ``os_primary`` Tempest credentials.

        Temporarily change the role used by ``os_primary`` credentials to:

        * ``[patrole] rbac_test_roles`` before test execution
        * ``[identity] admin_role`` after test execution

        Automatically switches to admin role after test execution.

        :returns: None

        .. warning::

            This function can alter user roles for pre-provisioned credentials.
            Work is underway to safely clean up after this function.

        Example::

            @rbac_rule_validation.action(service='test',
                                         rules=['a:test:rule'])
            def test_foo(self):
                # Allocate test-level resources here.
                with self.override_role():
                    # The role for `os_primary` has now been overridden. Within
                    # this block, call the API endpoint that enforces the
                    # expected policy specified by "rule" in the decorator.
                    self.foo_service.bar_api_call()
                # The role is switched back to admin automatically. Note that
                # if the API call above threw an exception, any code below this
                # point in the test is not executed.
        """
        self._set_override_role_called()
        self._override_role(True)
        try:
            # Execute the test.
            yield
        finally:
            # Check whether an exception was raised. If so, remember that
            # for future validation.
            exc = sys.exc_info()[0]
            if exc is not None:
                self._set_override_role_caught_exc()
            # This code block is always executed, no matter the result of the
            # test. Automatically switch back to the admin role for test clean
            # up.
            self._override_role(False)

    @classmethod
    def _override_role(cls, toggle_rbac_role=False):
        """Private helper for overriding ``os_primary`` Tempest credentials.

        :param toggle_rbac_role: Boolean value that controls the role that
            overrides default role of ``os_primary`` credentials.
            * If True: role is set to ``[patrole] rbac_test_role``
            * If False: role is set to ``[identity] admin_role``
        """
        LOG.debug('Overriding role to: %s.', toggle_rbac_role)
        roles_already_present = False

        try:
            target_roles = (cls._rbac_role_ids
                            if toggle_rbac_role else [cls._admin_role_id])
            roles_already_present = cls._list_and_clear_user_roles_on_project(
                target_roles)

            # Do not override roles if `target_role` already exists.
            if not roles_already_present:
                cls._create_user_role_on_project(target_roles)
        except Exception as exp:
            with excutils.save_and_reraise_exception():
                LOG.exception(exp)
        finally:
            auth_providers = cls.get_auth_providers()
            for provider in auth_providers:
                provider.clear_auth()
            # Fernet tokens are not subsecond aware so sleep to ensure we are
            # passing the second boundary before attempting to authenticate.
            # Only sleep if a token revocation occurred as a result of role
            # overriding. This will optimize test runtime in the case where
            # ``[identity] admin_role`` == ``[patrole] rbac_test_roles``.
            if not roles_already_present:
                time.sleep(1)

            for provider in auth_providers:
                provider.set_auth()

    @classmethod
    def _init_roles(cls):
        available_roles = cls.admin_roles_client.list_roles()['roles']
        cls._role_map = {r['name']: r['id'] for r in available_roles}
        LOG.debug('Available roles: %s', cls._role_map.keys())

        rbac_role_ids = []
        roles = CONF.patrole.rbac_test_roles
        # TODO(vegasq) drop once CONF.patrole.rbac_test_role is removed
        if CONF.patrole.rbac_test_role:
            if not roles:
                roles.append(CONF.patrole.rbac_test_role)

        for role_name in roles:
            rbac_role_ids.append(cls._role_map.get(role_name))

        admin_role_id = cls._role_map.get(CONF.identity.admin_role)

        if not all([admin_role_id, all(rbac_role_ids)]):
            missing_roles = []
            msg = ("Could not find `[patrole] rbac_test_roles` or "
                   "`[identity] admin_role`, both of which are required for "
                   "RBAC testing.")
            if not admin_role_id:
                missing_roles.append(CONF.identity.admin_role)
            if not all(rbac_role_ids):
                missing_roles += [role_name for role_name in roles
                                  if role_name not in cls._role_map]

            msg += " Following roles were not found: %s." % (
                ", ".join(missing_roles))
            msg += " Available roles: %s." % ", ".join(cls._role_map)
            raise rbac_exceptions.RbacResourceSetupFailed(msg)

        cls._admin_role_id = admin_role_id
        cls._rbac_role_ids = rbac_role_ids
        # Adding backward mapping
        cls._role_map.update({v: k for k, v in cls._role_map.items()})

    @classmethod
    def _create_user_role_on_project(cls, role_ids):
        for role_id in role_ids:
            cls.admin_roles_client.create_user_role_on_project(
                cls._project_id, cls._user_id, role_id)

    @classmethod
    def _list_and_clear_user_roles_on_project(cls, role_ids):
        roles = cls.admin_roles_client.list_user_roles_on_project(
            cls._project_id, cls._user_id)['roles']
        all_role_ids = [role['id'] for role in roles]

        # NOTE(felipemonteiro): We do not use ``role_id in all_role_ids`` here
        # to avoid over-permission errors: if the current list of roles on the
        # project includes "admin" and "Member", and we are switching to the
        # "Member" role, then we must delete the "admin" role. Thus, we only
        # return early if the user's roles on the project are an exact match.
        if set(role_ids) == set(all_role_ids):
            return True

        for role in roles:
            cls.admin_roles_client.delete_role_from_user_on_project(
                cls._project_id, cls._user_id, role['id'])

        return False

    @contextlib.contextmanager
    def override_role_and_validate_list(self,
                                        admin_resources=None,
                                        admin_resource_id=None):
        """Call ``override_role`` and validate RBAC for a list API action.

        List actions usually do soft authorization: partial or empty response
        bodies are returned instead of exceptions. This helper validates
        that unauthorized roles only return a subset of the available
        resources.
        Should only be used for validating list API actions.

        :param test_obj: Instance of ``tempest.test.BaseTestCase``.
        :param list admin_resources: The list of resources received before
            calling the ``override_role_and_validate_list`` function.
        :param UUID admin_resource_id: An ID of a resource created before
            calling the ``override_role_and_validate_list`` function.
        :return: py:class:`_ValidateListContext` object.

        Example::

            # the resource created by admin
            admin_resource_id = (
                self.ntp_client.create_dscp_marking_rule()
                ["dscp_marking_rule"]["id'])
            with self.override_role_and_validate_list(
                    admin_resource_id=admin_resource_id) as ctx:
                # the list of resources available for member role
                ctx.resources = self.ntp_client.list_dscp_marking_rules(
                    policy_id=self.policy_id)["dscp_marking_rules"]
        """
        ctx = _ValidateListContext(admin_resources, admin_resource_id)
        with self.override_role():
            yield ctx
            ctx._validate()

    @classmethod
    def get_auth_providers(cls):
        """Returns list of auth_providers used within test.

        Tests may redefine this method to include their own or third party
        client auth_providers.
        """
        return [cls.os_primary.auth_provider]

    def _set_override_role_called(self):
        """Helper for tracking whether ``override_role`` was called."""
        self.__override_role_called = True

    def _set_override_role_caught_exc(self):
        """Helper for tracking whether exception was thrown inside
        ``override_role``.
        """
        self.__override_role_caught_exc = True

    def _validate_override_role_called(self):
        """Idempotently validate that ``override_role`` is called and reset
        its value to False for sequential tests.
        """
        was_called = self.__override_role_called
        self.__override_role_called = False
        return was_called

    def _validate_override_role_caught_exc(self):
        """Idempotently validate that exception was caught inside
        ``override_role``, so that, by process of elimination, it can be
        determined whether one was thrown outside (which is invalid).
        """
        caught_exception = self.__override_role_caught_exc
        self.__override_role_caught_exc = False
        return caught_exception


def is_admin():
    """Verifies whether the current test role equals the admin role.

    :returns: True if ``rbac_test_roles`` contain the admin role.
    """
    roles = CONF.patrole.rbac_test_roles
    # TODO(vegasq) drop once CONF.patrole.rbac_test_role is removed
    if CONF.patrole.rbac_test_role:
        roles.append(CONF.patrole.rbac_test_role)
        roles = list(set(roles))

    # TODO(felipemonteiro): Make this more robust via a context is admin
    # lookup.
    return CONF.identity.admin_role in roles



# Copyright 2018 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.network import rbac_base as base


class RbacPoliciesExtRbacTest(base.BaseNetworkExtRbacTest):

    @classmethod
    def resource_setup(cls):
        super(RbacPoliciesExtRbacTest, cls).resource_setup()
        cls.tenant_id = cls.os_primary.credentials.tenant_id
        cls.network_id = cls.create_network()['id']

    def create_rbac_policy(self, tenant_id, network_id):
        policy = self.ntp_client.create_rbac_policy(
            target_tenant=tenant_id,
            object_type="network",
            object_id=network_id,
            action="access_as_shared"
        )
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            self.ntp_client.delete_rbac_policy, policy["rbac_policy"]["id"])

        return policy["rbac_policy"]["id"]

    @decorators.idempotent_id('effd9545-99ad-4c3c-92dd-ea422602c868')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["create_rbac_policy",
                                        "create_rbac_policy:target_tenant"])
    def test_create_rbac_policy(self):
        """Create RBAC policy.

        RBAC test for the neutron "create_rbac_policy" policy

        We can't validate "create_rbac_policy:target_tenant" for all cases
        since if "restrict_wildcard" rule is modified then Patrole won't be
        able to determine the correct result since that requires relying on
        Neutron's custom FieldCheck oslo.policy rule.
        """

        with self.override_role():
            self.create_rbac_policy(self.tenant_id, self.network_id)

    @decorators.idempotent_id('f5d836d8-3b64-412d-a283-ee29761017f3')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_rbac_policy",
                                        "update_rbac_policy",
                                        "update_rbac_policy:target_tenant"],
                                 expected_error_codes=[404, 403, 403])
    def test_update_rbac_policy(self):
        """Update RBAC policy.

        RBAC test for the neutron "update_rbac_policy" policy

        We can't validate "create_rbac_policy:target_tenant" for all cases
        since if "restrict_wildcard" rule is modified then Patrole won't be
        able to determine the correct result since that requires relying on
        Neutron's custom FieldCheck oslo.policy rule.
        """
        policy_id = self.create_rbac_policy(self.tenant_id, self.network_id)

        with self.override_role():
            self.ntp_client.update_rbac_policy(
                policy_id, target_tenant=self.tenant_id)

    @decorators.idempotent_id('9308ab18-426c-41b7-bce5-11081f7dd259')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_rbac_policy"],
                                 expected_error_codes=[404])
    def test_show_rbac_policy(self):
        """Show RBAC policy.

        RBAC test for the neutron "get_rbac_policy" policy
        """
        policy_id = self.create_rbac_policy(self.tenant_id, self.network_id)

        with self.override_role():
            self.ntp_client.show_rbac_policy(policy_id)

    @decorators.idempotent_id('54aa9bce-efea-47fb-b0e4-12012f82f285')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_rbac_policy",
                                        "delete_rbac_policy"],
                                 expected_error_codes=[404, 403])
    def test_delete_rbac_policy(self):
        """Delete RBAC policy.

        RBAC test for the neutron "delete_rbac_policy" policy
        """
        policy_id = self.create_rbac_policy(self.tenant_id, self.network_id)

        with self.override_role():
            self.ntp_client.delete_rbac_policy(policy_id)

    @decorators.idempotent_id('5337d95a-2e75-47bb-a0ea-0a082be930bf')
    @rbac_rule_validation.action(service="neutron", rules=["get_rbac_policy"])
    def test_list_rbac_policies(self):
        """List RBAC policies.

        RBAC test for the neutron ``list_rbac_policies`` function and
        the ``get_rbac_policy`` policy
        """
        admin_resource_id = self.create_rbac_policy(self.tenant_id,
                                                    self.network_id)
        with (self.override_role_and_validate_list(
                admin_resource_id=admin_resource_id)) as ctx:
            ctx.resources = self.ntp_client.list_rbac_policies(
                id=admin_resource_id)["rbac_policies"]



#    Copyright 2017 AT&T Corporation.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tempest.common import utils
from tempest import config
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base

CONF = config.CONF


class FlavorRxtxRbacTest(rbac_base.BaseV2ComputeRbacTest):

    @classmethod
    def skip_checks(cls):
        super(FlavorRxtxRbacTest, cls).skip_checks()
        if not utils.is_extension_enabled('os-flavor-rxtx', 'compute'):
            msg = "os-flavor-rxtx extension not enabled."
            raise cls.skipException(msg)

    @testtools.skipIf(CONF.policy_feature_enabled.removed_nova_policies_stein,
                      "This API extension policy was removed in Stein")
    @decorators.idempotent_id('5e1fd9f0-9a08-485a-ad9c-0fc66e4d64b7')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-flavor-rxtx"])
    def test_list_flavors_details_rxtx(self):
        with self.override_role():
            result = self.flavors_client.list_flavors(detail=True)['flavors']
        if 'rxtx_factor' not in result[0]:
            raise rbac_exceptions.RbacMissingAttributeResponseBody(
                attribute='rxtx_factor')

    @testtools.skipIf(CONF.policy_feature_enabled.removed_nova_policies_stein,
                      "This API extension policy was removed in Stein")
    @decorators.idempotent_id('70c55a07-c843-4627-a29d-ba78673c1e63')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-flavor-rxtx"])
    def test_get_flavor_rxtx(self):
        with self.override_role():
            result = self.flavors_client.show_flavor(
                CONF.compute.flavor_ref)['flavor']
        if 'rxtx_factor' not in result:
            raise rbac_exceptions.RbacMissingAttributeResponseBody(
                attribute='rxtx_factor')



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.volume import rbac_base

CONF = config.CONF


class SnapshotMetadataV3RbacTest(rbac_base.BaseVolumeRbacTest):

    @classmethod
    def skip_checks(cls):
        super(SnapshotMetadataV3RbacTest, cls).skip_checks()
        if not CONF.volume_feature_enabled.snapshot:
            raise cls.skipException("Cinder snapshot feature disabled")

    @classmethod
    def resource_setup(cls):
        super(SnapshotMetadataV3RbacTest, cls).resource_setup()
        cls.volume = cls.create_volume()
        # Create a snapshot
        cls.snapshot = cls.create_snapshot(volume_id=cls.volume['id'])
        cls.snapshot_id = cls.snapshot['id']

    @classmethod
    def _create_test_snapshot_metadata(cls):
        # Create test snapshot metadata
        metadata = {"key1": "value1",
                    "key2": "value2",
                    "key3": "value3"}
        cls.snapshots_client.create_snapshot_metadata(
            cls.snapshot_id, metadata)['metadata']

    @rbac_rule_validation.action(service="cinder",
                                 rules=["volume:get_snapshot_metadata"])
    @decorators.idempotent_id('f6912bb1-62e6-483d-bcd0-e98c1641f4c3')
    def test_get_snapshot_metadata(self):
        # Create volume and snapshot metadata
        self._create_test_snapshot_metadata()
        # Get metadata for the snapshot
        with self.override_role():
            self.snapshots_client.show_snapshot_metadata(
                self.snapshot_id)

    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:volume_tenant_attribute"])
    @decorators.idempotent_id('e2c73b00-0c19-4bb7-8c61-d84b1a223ed1')
    def test_get_snapshot_metadata_for_volume_tenant(self):
        # Create volume and snapshot metadata
        self._create_test_snapshot_metadata()
        # Get the metadata of the snapshot
        with self.override_role():
            self.snapshots_client.show_snapshot_metadata(
                self.snapshot_id)['metadata']

    @decorators.idempotent_id('7ea597f6-c544-4b10-aab0-ff68f595fb06')
    @rbac_rule_validation.action(service="cinder",
                                 rules=["volume:update_snapshot_metadata"])
    def test_update_snapshot_metadata(self):
        self._create_test_snapshot_metadata()
        with self.override_role():
            update = {"key3": "value3_update",
                      "key4": "value4"}
            self.snapshots_client.update_snapshot_metadata(
                self.snapshot['id'], metadata=update)

    @decorators.idempotent_id('93068d02-0131-4dd3-af16-fc40d7128d93')
    @rbac_rule_validation.action(service="cinder",
                                 rules=["volume:get_snapshot_metadata"])
    def test_show_snapshot_metadata_item(self):
        self._create_test_snapshot_metadata()
        with self.override_role():
            self.snapshots_client.show_snapshot_metadata_item(
                self.snapshot['id'], "key3")['meta']

    @decorators.idempotent_id('1f8f43e7-da31-4128-bb3c-73fc548650e3')
    @rbac_rule_validation.action(service="cinder",
                                 rules=["volume:update_snapshot_metadata"])
    def test_update_snapshot_metadata_item(self):
        update_item = {"key3": "value3_update"}
        self._create_test_snapshot_metadata()
        with self.override_role():
            self.snapshots_client.update_snapshot_metadata_item(
                self.snapshot['id'], "key3", meta=update_item)['meta']

    @decorators.idempotent_id('3ec32516-f7cd-4f88-b78a-ddee67492071')
    @rbac_rule_validation.action(service="cinder",
                                 rules=["volume:delete_snapshot_metadata"])
    def test_delete_snapshot_metadata_item(self):
        self._create_test_snapshot_metadata()
        with self.override_role():
            self.snapshots_client.delete_snapshot_metadata_item(
                self.snapshot['id'], "key1")



#    Copyright 2017 AT&T Corporation.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base

CONF = config.CONF


# TODO(rb560u): Remove this test class once the nova queens branch goes into
# extended maintenance mode.
class VirtualInterfacesRbacTest(rbac_base.BaseV2ComputeRbacTest):
    # The compute os-virtual-interfaces API is deprecated from the Microversion
    # 2.44 onward. For more information, see:
    # https://developer.openstack.org/api-ref/compute/#servers-virtual-interfaces-servers-os-virtual-interfaces-deprecated
    depends_on_nova_network = True
    max_microversion = '2.43'

    @classmethod
    def setup_credentials(cls):
        # This test needs a network and a subnet
        cls.set_network_resources(network=True, subnet=True)
        super(VirtualInterfacesRbacTest, cls).setup_credentials()

    @classmethod
    def resource_setup(cls):
        super(VirtualInterfacesRbacTest, cls).resource_setup()
        cls.server = cls.create_test_server(wait_until='ACTIVE')

    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-virtual-interfaces"])
    @decorators.idempotent_id('fc719ae3-0f73-4689-8378-1b841f0f2818')
    def test_list_virtual_interfaces(self):
        """Test list virtual interfaces, part of os-virtual-interfaces.

        If Neutron is available, then call the API and expect it to fail
        with a 400 BadRequest (policy enforcement is done before that happens).
        """
        with self.override_role():
            if CONF.service_available.neutron:
                msg = ("Listing virtual interfaces is not supported by this "
                       "cloud.")
                with self.assertRaisesRegex(lib_exc.BadRequest, msg):
                    self.servers_client.list_virtual_interfaces(
                        self.server['id'])
            else:
                self.servers_client.list_virtual_interfaces(self.server['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.volume import rbac_base

CONF = config.CONF


class MessagesV3RbacTest(rbac_base.BaseVolumeRbacTest):
    min_microversion = '3.3'
    max_microversion = 'latest'

    @classmethod
    def setup_clients(cls):
        super(MessagesV3RbacTest, cls).setup_clients()
        cls.messages_client = cls.os_primary.volume_messages_client_latest

    def _create_user_message(self):
        """Trigger a 'no valid host' situation to generate a message."""
        bad_protocol = data_utils.rand_name(
            self.__class__.__name__ + '-storage_protocol')
        bad_vendor = data_utils.rand_name(
            self.__class__.__name__ + '-vendor_name')
        extra_specs = {'storage_protocol': bad_protocol,
                       'vendor_name': bad_vendor}
        vol_type_name = data_utils.rand_name(
            self.__class__.__name__ + '-volume-type')
        bogus_type = self.create_volume_type(
            name=vol_type_name, extra_specs=extra_specs)
        params = {'volume_type': bogus_type['id'],
                  'size': CONF.volume.volume_size}
        volume = self.create_volume(wait_until="error", **params)
        messages = self.messages_client.list_messages()['messages']
        message_id = None
        for message in messages:
            if message['resource_uuid'] == volume['id']:
                message_id = message['id']
                break
        self.assertIsNotNone(message_id, 'No user message generated for '
                                         'volume %s' % volume['id'])

        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.messages_client.delete_message, message_id)

        return message_id

    @decorators.idempotent_id('bf7f31a1-509b-4a7d-a8a8-ad6ce68229c7')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["message:get_all"])
    def test_list_messages(self):
        with self.override_role():
            self.messages_client.list_messages()['messages']

    @decorators.idempotent_id('9cc1ad1e-68a2-4407-8b60-ea77909bce08')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["message:get"])
    def test_show_message(self):
        message_id = self._create_user_message()

        with self.override_role():
            self.messages_client.show_message(message_id)['message']

    @decorators.idempotent_id('65ca7fb7-7f2c-443e-b144-ac86973a97be')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["message:delete"])
    def test_delete_message(self):
        message_id = self._create_user_message()

        with self.override_role():
            self.messages_client.delete_message(message_id)
        self.messages_client.wait_for_resource_deletion(message_id)



# Copyright 2018 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.network import rbac_base as base


class FlavorsServiceProfileExtRbacTest(base.BaseNetworkExtRbacTest):
    @classmethod
    def resource_setup(cls):
        super(FlavorsServiceProfileExtRbacTest, cls).resource_setup()
        providers = cls.ntp_client.list_service_providers()
        if not providers["service_providers"]:
            raise cls.skipException("No service_providers available.")
        cls.service_type = providers["service_providers"][0]["service_type"]

        cls.flavor_id = cls.create_flavor()
        cls.service_profile_id = cls.create_service_profile()

    @classmethod
    def create_flavor(cls):
        flavor = cls.ntp_client.create_flavor(service_type=cls.service_type)
        flavor_id = flavor["flavor"]["id"]
        cls.addClassResourceCleanup(
            test_utils.call_and_ignore_notfound_exc,
            cls.ntp_client.delete_flavor, flavor_id)
        return flavor_id

    def create_flavor_service_profile(self, flavor_id, service_profile_id):
        self.ntp_client.create_flavor_service_profile(
            flavor_id, service_profile_id)
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            self.ntp_client.delete_flavor_service_profile,
            flavor_id, service_profile_id)

    @decorators.idempotent_id('aa84b4c5-0dd6-4c34-aa81-3a76507f9b81')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["create_flavor_service_profile"])
    def test_create_flavor_service_profile(self):
        """Create flavor_service_profile.

        RBAC test for the neutron "create_flavor_service_profile" policy
        """
        with self.override_role():
            self.create_flavor_service_profile(self.flavor_id,
                                               self.service_profile_id)

    @decorators.idempotent_id('3b680d9e-946a-4670-ab7f-0e4576675833')
    @rbac_rule_validation.action(service="neutron",
                                 rules=["delete_flavor_service_profile"])
    def test_delete_flavor_service_profile(self):
        """Delete flavor_service_profile.

        RBAC test for the neutron "delete_flavor_service_profile" policy
        """
        self.create_flavor_service_profile(self.flavor_id,
                                           self.service_profile_id)

        with self.override_role():
            self.ntp_client.delete_flavor_service_profile(
                self.flavor_id, self.service_profile_id)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.volume import rbac_base


class VolumeTypesRbacTest(rbac_base.BaseVolumeRbacTest):

    @decorators.idempotent_id('e2bbf968-d947-4a15-a4da-a98c3069731e')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:types_manage"])
    def test_create_volume_type(self):
        with self.override_role():
            self.create_volume_type()

    @decorators.idempotent_id('2b74ac82-e03e-4801-86f3-d05c9acfd66b')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:types_manage"])
    def test_update_volume_type(self):
        volume_type = self.create_volume_type()
        with self.override_role():
            self.volume_types_client.update_volume_type(
                volume_type['id'], description='updated-description')

    @decorators.idempotent_id('90aec0ef-4f9b-4170-be6b-a392c12540be')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:types_manage"])
    def test_delete_volume_type(self):
        volume_type = self.create_volume_type()
        with self.override_role():
            self.volume_types_client.delete_volume_type(volume_type['id'])
        self.volume_types_client.wait_for_resource_deletion(volume_type['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import rbac_utils
from patrole_tempest_plugin.tests.unit import base
from patrole_tempest_plugin.tests.unit import fixtures as patrole_fixtures


class RBACUtilsMixinTest(base.TestCase):

    def setUp(self):
        super(RBACUtilsMixinTest, self).setUp()
        self.rbac_utils_fixture = self.useFixture(
            patrole_fixtures.RbacUtilsMixinFixture())
        self.test_obj = self.rbac_utils_fixture.test_obj

    def test_init_roles_with_missing_admin_role(self):
        self.rbac_utils_fixture.set_roles('member')
        error_re = (".*Following roles were not found: admin. Available "
                    "roles: member.")
        self.assertRaisesRegex(rbac_exceptions.RbacResourceSetupFailed,
                               error_re, self.test_obj._init_roles)

    def test_init_roles_with_missing_rbac_role(self):
        self.rbac_utils_fixture.set_roles('admin')
        error_re = (".*Following roles were not found: member. Available "
                    "roles: admin.")
        self.assertRaisesRegex(rbac_exceptions.RbacResourceSetupFailed,
                               error_re, self.test_obj._init_roles)

    def test_override_role_to_admin_role_at_creating(self):
        rbac_utils_fixture = self.useFixture(
            patrole_fixtures.RbacUtilsMixinFixture(do_reset_mocks=False))
        test_obj = rbac_utils_fixture.test_obj
        roles_client = rbac_utils_fixture.admin_roles_client
        mock_time = rbac_utils_fixture.mock_time

        roles_client.create_user_role_on_project.assert_called_once_with(
            rbac_utils_fixture.PROJECT_ID,
            rbac_utils_fixture.USER_ID,
            'admin_id')
        test_obj.get_auth_providers()[0].clear_auth.assert_called_once_with()
        test_obj.get_auth_providers()[0].set_auth.assert_called_once_with()
        mock_time.sleep.assert_called_once_with(1)

    def test_override_role_to_admin_role(self):
        self.test_obj._override_role()

        roles_client = self.rbac_utils_fixture.admin_roles_client
        mock_time = self.rbac_utils_fixture.mock_time

        roles_client.create_user_role_on_project.assert_called_once_with(
            self.rbac_utils_fixture.PROJECT_ID,
            self.rbac_utils_fixture.USER_ID,
            'admin_id')
        self.test_obj.get_auth_providers()[0].clear_auth\
            .assert_called_once_with()
        self.test_obj.get_auth_providers()[0].set_auth\
            .assert_called_once_with()
        mock_time.sleep.assert_called_once_with(1)

    def test_override_role_to_admin_role_avoids_role_switch(self):
        self.rbac_utils_fixture.set_roles(['admin', 'member'], 'admin')
        self.test_obj._override_role()

        roles_client = self.rbac_utils_fixture.admin_roles_client
        mock_time = self.rbac_utils_fixture.mock_time

        roles_client.create_user_role_on_project.assert_not_called()
        mock_time.sleep.assert_not_called()

    def test_override_role_to_member_role(self):
        self.test_obj._override_role(True)

        roles_client = self.rbac_utils_fixture.admin_roles_client
        mock_time = self.rbac_utils_fixture.mock_time

        roles_client.create_user_role_on_project.assert_has_calls([
            mock.call(self.rbac_utils_fixture.PROJECT_ID,
                      self.rbac_utils_fixture.USER_ID,
                      'member_id')
        ])
        self.test_obj.get_auth_providers()[0].clear_auth.assert_has_calls(
            [mock.call()])
        self.test_obj.get_auth_providers()[0].set_auth.assert_has_calls(
            [mock.call()])
        mock_time.sleep.assert_has_calls([mock.call(1)])

    def test_override_role_to_member_role_avoids_role_switch(self):
        self.rbac_utils_fixture.set_roles(['admin', 'member'], 'member')
        self.test_obj._override_role(True)

        roles_client = self.rbac_utils_fixture.admin_roles_client
        mock_time = self.rbac_utils_fixture.mock_time

        self.assertEqual(0,
                         roles_client.create_user_role_on_project.call_count)
        self.assertEqual(0,
                         mock_time.sleep.call_count)

    def test_override_role_to_member_role_then_admin_role(self):
        self.test_obj._override_role(True)
        self.test_obj._override_role(False)

        roles_client = self.rbac_utils_fixture.admin_roles_client
        mock_time = self.rbac_utils_fixture.mock_time

        roles_client.create_user_role_on_project.assert_has_calls([
            mock.call(self.rbac_utils_fixture.PROJECT_ID,
                      self.rbac_utils_fixture.USER_ID,
                      'member_id'),
            mock.call(self.rbac_utils_fixture.PROJECT_ID,
                      self.rbac_utils_fixture.USER_ID,
                      'admin_id')
        ])
        self.test_obj.get_auth_providers()[0].clear_auth.assert_has_calls(
            [mock.call()] * 2)
        self.test_obj.get_auth_providers()[0].set_auth.assert_has_calls(
            [mock.call()] * 2)
        mock_time.sleep.assert_has_calls([mock.call(1)] * 2)

    def test_clear_user_roles(self):
        # NOTE(felipemonteiro): Set the user's roles on the project to
        # include 'random' to coerce a role switch, or else it will be
        # skipped.
        self.rbac_utils_fixture.set_roles(['admin', 'member'],
                                          ['member', 'random'])
        self.test_obj._override_role()

        roles_client = self.rbac_utils_fixture.admin_roles_client

        roles_client.list_user_roles_on_project.assert_called_once_with(
            self.rbac_utils_fixture.PROJECT_ID,
            self.rbac_utils_fixture.USER_ID)
        roles_client.delete_role_from_user_on_project.\
            assert_has_calls([
                mock.call(mock.sentinel.project_id, mock.sentinel.user_id,
                          'member_id'),
                mock.call(mock.sentinel.project_id, mock.sentinel.user_id,
                          'random_id')])

    def test_override_role_context_manager_simulate_pass(self):
        """Validate that expected override_role calls are made when switching
        to admin role for success path.
        """

        mock_override_role = self.patchobject(self.test_obj, '_override_role')
        with self.test_obj.override_role():
            # Validate `override_role` public method called private method
            # `_override_role` with True.
            mock_override_role.assert_called_once_with(True)
            mock_override_role.reset_mock()
        # Validate that `override_role` switched back to admin role after
        # contextmanager.
        mock_override_role.assert_called_once_with(False)

    def test_override_role_context_manager_simulate_fail(self):
        """Validate that expected override_role calls are made when switching
        to admin role for failure path (i.e. when test raises exception).
        """
        mock_override_role = self.patchobject(self.test_obj, '_override_role')

        def _do_test():
            with self.test_obj.override_role():
                # Validate `override_role` public method called private method
                # `_override_role` with True.
                mock_override_role.assert_called_once_with(True)
                mock_override_role.reset_mock()
                # Raise exc to verify role switch works for negative case.
                raise lib_exc.Forbidden()

        # Validate that role is switched back to admin, despite test failure.
        with testtools.ExpectedException(lib_exc.Forbidden):
            _do_test()
        mock_override_role.assert_called_once_with(False)

    def test_override_role_and_validate_list(self):
        m_override_role = self.patchobject(self.test_obj, 'override_role')

        with (self.test_obj.override_role_and_validate_list(
                admin_resource_id='foo')) as ctx:
            self.assertIsInstance(ctx, rbac_utils._ValidateListContext)
            m_validate = self.patchobject(ctx, '_validate')
        m_override_role.assert_called_once_with()
        m_validate.assert_called_once()

    def test_prepare_role_inferences_mapping(self):
        self.test_obj.admin_roles_client.list_all_role_inference_rules.\
            return_value = {
                "role_inferences": [
                    {
                        "implies": [{"id": "reader_id", "name": "reader"}],
                        "prior_role": {"id": "member_id", "name": "member"}
                    },
                    {
                        "implies": [{"id": "member_id", "name": "member"}],
                        "prior_role": {"id": "admin_id", "name": "admin"}
                    }
                ]
            }

        expected_role_inferences_mapping = {
            "member_id": {"reader_id"},
            "admin_id": {"member_id", "reader_id"}
        }
        actual_role_inferences_mapping = self.test_obj.\
            _prepare_role_inferences_mapping()
        self.assertEqual(expected_role_inferences_mapping,
                         actual_role_inferences_mapping)

    def test_get_all_needed_roles(self):
        self.test_obj.__class__._role_inferences_mapping = {
            "member_id": {"reader_id"},
            "admin_id": {"member_id", "reader_id"}
        }
        self.test_obj.__class__._role_map = {
            "admin_id": "admin", "admin": "admin_id",
            "member_id": "member", "member": "member_id",
            "reader_id": "reader", "reader": "reader_id"
        }
        for roles, expected_roles in (
            (['admin'], ['admin', 'member', 'reader']),
            (['member'], ['member', 'reader']),
            (['reader'], ['reader']),
            (['custom_role'], ['custom_role']),
            (['custom_role', 'member'], ['custom_role', 'member', 'reader']),
            (['admin', 'member'], ['admin', 'member', 'reader']),
        ):
            expected_roles = sorted(expected_roles)
            actual_roles = sorted(self.test_obj.get_all_needed_roles(roles))
        self.assertEqual(expected_roles, actual_roles)


class ValidateListContextTest(base.TestCase):
    @staticmethod
    def _get_context(admin_resources=None, admin_resource_id=None):
        return rbac_utils._ValidateListContext(
            admin_resources=admin_resources,
            admin_resource_id=admin_resource_id)

    def test_incorrect_usage(self):
        # admin_resources and admin_resource_is are not assigned
        self.assertRaises(rbac_exceptions.RbacValidateListException,
                          self._get_context)

        # both admin_resources and admin_resource_is are assigned
        self.assertRaises(rbac_exceptions.RbacValidateListException,
                          self._get_context,
                          admin_resources='foo', admin_resource_id='bar')
        # empty list assigned to admin_resources
        self.assertRaises(rbac_exceptions.RbacValidateListException,
                          self._get_context, admin_resources=[])

        # ctx.resources is not assigned
        ctx = self._get_context(admin_resources='foo')
        self.assertRaises(rbac_exceptions.RbacValidateListException,
                          ctx._validate)

    def test_validate_len_negative(self):
        ctx = self._get_context(admin_resources=[1, 2, 3, 4])
        self.assertEqual(ctx._validate_len, ctx._validate_func)
        self.assertEqual(4, ctx._admin_len)
        self.assertFalse(hasattr(ctx, '_admin_resource_id'))

        # the number of resources is less than admin resources
        ctx.resources = [1, 2, 3]
        self.assertRaises(rbac_exceptions.RbacPartialResponseBody,
                          ctx._validate_len)

        # the resources is empty
        ctx.resources = []
        self.assertRaises(rbac_exceptions.RbacEmptyResponseBody,
                          ctx._validate_len)

    def test_validate_len(self):
        ctx = self._get_context(admin_resources=[1, 2, 3, 4])

        # the number of resources and admin resources are same
        ctx.resources = [1, 2, 3, 4]
        self.assertIsNone(ctx._validate_len())

    def test_validate_resource_negative(self):
        ctx = self._get_context(admin_resource_id=1)
        self.assertEqual(ctx._validate_resource, ctx._validate_func)
        self.assertEqual(1, ctx._admin_resource_id)
        self.assertFalse(hasattr(ctx, '_admin_len'))

        # there is no admin resource in the resources
        ctx.resources = [{'id': 2}, {'id': 3}]
        self.assertRaises(rbac_exceptions.RbacPartialResponseBody,
                          ctx._validate_resource)

    def test_validate_resource(self):
        ctx = self._get_context(admin_resource_id=1)

        # there is admin resource in the resources
        ctx.resources = [{'id': 1}, {'id': 2}]
        self.assertIsNone(ctx._validate_resource())

    def test_validate(self):
        ctx = self._get_context(admin_resources='foo')
        ctx.resources = 'bar'
        with mock.patch.object(ctx, '_validate_func',
                               autospec=False) as m_validate_func:
            m_validate_func.side_effect = (
                rbac_exceptions.RbacPartialResponseBody,
                None
            )
            self.assertRaises(rbac_exceptions.RbacPartialResponseBody,
                              ctx._validate)
            m_validate_func.assert_called_once()

            m_validate_func.reset_mock()
            ctx._validate()
            m_validate_func.assert_called_once()



#    Copyright 2017 AT&T Corporation.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base


class HostsRbacTest(rbac_base.BaseV2ComputeRbacTest):
    # These tests will fail with a 404 starting from microversion 2.43:
    # See the following links for details:
    # https://developer.openstack.org/api-ref/compute/#hosts-os-hosts-deprecated
    max_microversion = '2.42'

    @classmethod
    def skip_checks(cls):
        super(HostsRbacTest, cls).skip_checks()
        if not utils.is_extension_enabled('os-hosts', 'compute'):
            msg = "%s skipped as os-hosts not enabled." % cls.__name__
            raise cls.skipException(msg)

    @decorators.idempotent_id('035b7935-2fae-4218-8d37-27fa83097494')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-hosts"])
    def test_list_hosts(self):
        with self.override_role():
            self.hosts_client.list_hosts()

    @decorators.idempotent_id('bc10d8b4-d2c3-4d4e-9d2b-31d1bd3e1b51')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-hosts"])
    def test_show_host_details(self):
        hosts = self.hosts_client.list_hosts()['hosts']
        hosts = [host for host in hosts if host['service'] == 'compute']
        self.assertNotEmpty(hosts)

        with self.override_role():
            self.hosts_client.show_host(hosts[0]['host_name'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.identity import rbac_base


class EndpointFilterProjectsV3RbacTest(rbac_base.BaseIdentityV3RbacTest):

    @classmethod
    def resource_setup(cls):
        super(EndpointFilterProjectsV3RbacTest, cls).resource_setup()
        cls.project = cls.setup_test_project()
        cls.endpoint = cls.setup_test_endpoint()

    def _add_endpoint_to_project(self, ignore_not_found=False):
        self.endpoint_filter_client.add_endpoint_to_project(
            self.project['id'], self.endpoint['id'])

        if ignore_not_found:
            self.addCleanup(
                test_utils.call_and_ignore_notfound_exc,
                self.endpoint_filter_client.delete_endpoint_from_project,
                self.project['id'], self.endpoint['id'])
        else:
            self.addCleanup(
                self.endpoint_filter_client.delete_endpoint_from_project,
                self.project['id'], self.endpoint['id'])

    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:add_endpoint_to_project"])
    @decorators.idempotent_id('9199ec13-816d-4efe-b8b1-e1cd026b9747')
    def test_add_endpoint_to_project(self):
        # Adding endpoints to projects
        with self.override_role():
            self._add_endpoint_to_project(ignore_not_found=True)

    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:list_projects_for_endpoint"])
    @decorators.idempotent_id('f53dca42-ec8a-48e9-924b-0bbe6c99727f')
    def test_list_projects_for_endpoint(self):
        with self.override_role():
            self.endpoint_filter_client.list_projects_for_endpoint(
                self.endpoint['id'])

    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:check_endpoint_in_project"])
    @decorators.idempotent_id('0c1425eb-833c-4aa1-a21d-52ffa41fdc6a')
    def test_check_endpoint_in_project(self):
        self._add_endpoint_to_project()
        with self.override_role():
            self.endpoint_filter_client.check_endpoint_in_project(
                self.project['id'], self.endpoint['id'])

    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:list_endpoints_for_project"])
    @decorators.idempotent_id('5d86c659-c6ad-41e0-854e-3823e95c7cc2')
    def test_list_endpoints_in_project(self):
        with self.override_role():
            self.endpoint_filter_client.list_endpoints_in_project(
                self.project['id'])

    @rbac_rule_validation.action(
        service="keystone",
        rules=["identity:remove_endpoint_from_project"])
    @decorators.idempotent_id('b4e21c10-4f47-427b-9b8a-f5b5601adfda')
    def test_remove_endpoint_from_project(self):
        self._add_endpoint_to_project(ignore_not_found=True)
        with self.override_role():
            self.endpoint_filter_client.delete_endpoint_from_project(
                self.project['id'], self.endpoint['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.identity import rbac_base


class IdentityPoliciesV3RbacTest(rbac_base.BaseIdentityV3RbacTest):

    @rbac_rule_validation.action(service="keystone",
                                 rules=["identity:create_policy"])
    @decorators.idempotent_id('de2f7ecb-fbf0-41f3-abf4-b97b5e082fd5')
    def test_create_policy(self):
        with self.override_role():
            self.setup_test_policy()

    @rbac_rule_validation.action(service="keystone",
                                 rules=["identity:update_policy"])
    @decorators.idempotent_id('9cfed3c6-0b27-4d15-be67-e06e0cfb01b9')
    def test_update_policy(self):
        policy = self.setup_test_policy()
        updated_policy_type = data_utils.rand_name(
            self.__class__.__name__ + '-policy_type')

        with self.override_role():
            self.policies_client.update_policy(policy['id'],
                                               type=updated_policy_type)

    @rbac_rule_validation.action(service="keystone",
                                 rules=["identity:delete_policy"])
    @decorators.idempotent_id('dcd93f75-1e1b-4fbe-bee0-9c4c7b201735')
    def test_delete_policy(self):
        policy = self.setup_test_policy()

        with self.override_role():
            self.policies_client.delete_policy(policy['id'])

    @rbac_rule_validation.action(service="keystone",
                                 rules=["identity:get_policy"])
    @decorators.idempotent_id('d7e415c2-945a-4504-9571-0e2d0dd8594b')
    def test_show_policy(self):
        policy = self.setup_test_policy()

        with self.override_role():
            self.policies_client.show_policy(policy['id'])

    @rbac_rule_validation.action(service="keystone",
                                 rules=["identity:list_policies"])
    @decorators.idempotent_id('35a56161-4054-4237-8a78-7ce805dce202')
    def test_list_policies(self):
        with self.override_role():
            self.policies_client.list_policies()



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import logging
import sys

from oslo_log import versionutils
from oslo_utils import excutils
import six

from tempest import config
from tempest.lib import exceptions as lib_exc
from tempest import test

from patrole_tempest_plugin import policy_authority
from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import requirements_authority

CONF = config.CONF
LOG = logging.getLogger(__name__)

_SUPPORTED_ERROR_CODES = [403, 404]
_DEFAULT_ERROR_CODE = 403

RBACLOG = logging.getLogger('rbac_reporting')


def action(service,
           rules=None,
           expected_error_codes=None,
           extra_target_data=None):
    """A decorator for verifying OpenStack policy enforcement.

    A decorator which allows for positive and negative RBAC testing. Given:

    * an OpenStack service,
    * a policy action (``rule``) enforced by that service, and
    * the test roles defined by ``[patrole] rbac_test_roles``

    determines whether the test role has sufficient permissions to perform an
    API call that enforces the ``rule``.

    This decorator should only be applied to an instance or subclass of
    ``tempest.test.BaseTestCase``.

    The result from ``_is_authorized`` is used to determine the *expected*
    test result. The *actual* test result is determined by running the
    Tempest test this decorator applies to.

    Below are the following possibilities from comparing the *expected* and
    *actual* results:

    1) If *expected* is True and the test passes (*actual*), this is a success.
    2) If *expected* is True and the test fails (*actual*), this results in a
       ``RbacUnderPermissionException`` exception failure.
    3) If *expected* is False and the test passes (*actual*), this results in
       an ``RbacOverPermissionException`` exception failure.
    4) If *expected* is False and the test fails (*actual*), this is a success.

    As such, negative and positive testing can be applied using this decorator.

    :param str service: An OpenStack service. Examples: "nova" or "neutron".
    :param list rules: A list of policy actions defined in a policy file or in
        code. The rules are logical-ANDed together to derive the expected
        result. Also accepts list of callables that return a policy action.

        .. note::

            Patrole currently only supports custom JSON policy files.

    :type rules: list[str] or list[callable]
    :param list expected_error_codes: When the ``rules`` list parameter is
        used, then this list indicates the expected error code to use if one
        of the rules does not allow the role being tested. This list must
        coincide with and its elements remain in the same order as the rules
        in the rules list.

        Example::

            rules=["api_action1", "api_action2"]
            expected_error_codes=[404, 403]

        a) If api_action1 fails and api_action2 passes, then the expected
           error code is 404.
        b) if api_action2 fails and api_action1 passes, then the expected
           error code is 403.
        c) if both api_action1 and api_action2 fail, then the expected error
           code is the first error seen (404).

        If it is not passed, then it is defaulted to 403.

        .. warning::

            A 404 should not be provided *unless* the endpoint masks a
            ``Forbidden`` exception as a ``NotFound`` exception.

    :type expected_error_codes: list[int]
    :param dict extra_target_data: Dictionary, keyed with ``oslo.policy``
        generic check names, whose values are string literals that reference
        nested ``tempest.test.BaseTestCase`` attributes. Used by
        ``oslo.policy`` for performing matching against attributes that are
        sent along with the API calls. Example::

            extra_target_data={
                "target.token.user_id":
                "os_alt.auth_provider.credentials.user_id"
            })

    :raises RbacInvalidServiceException: If ``service`` is invalid.
    :raises RbacUnderPermissionException: For item (2) above.
    :raises RbacOverPermissionException: For item (3) above.
    :raises RbacExpectedWrongException: When a 403 is expected but a 404
        is raised instead or vice versa.

    Examples::

        @rbac_rule_validation.action(
            service="nova",
            rules=["os_compute_api:os-agents"])
        def test_list_agents_rbac(self):
            # The call to `override_role` is mandatory.
            with self.override_role():
                self.agents_client.list_agents()
    """

    if extra_target_data is None:
        extra_target_data = {}

    rules, expected_error_codes = _prepare_multi_policy(rules,
                                                        expected_error_codes)

    def decorator(test_func):
        roles = CONF.patrole.rbac_test_roles
        # TODO(vegasq) drop once CONF.patrole.rbac_test_role is removed
        if CONF.patrole.rbac_test_role:
            msg = ('CONF.patrole.rbac_test_role is deprecated in favor of '
                   'CONF.patrole.rbac_test_roles and will be removed in '
                   'future.')
            versionutils.report_deprecated_feature(LOG, msg)
            if not roles:
                roles.append(CONF.patrole.rbac_test_role)

        @functools.wraps(test_func)
        def wrapper(*args, **kwargs):
            if args and isinstance(args[0], test.BaseTestCase):
                test_obj = args[0]
            else:
                raise rbac_exceptions.RbacResourceSetupFailed(
                    '`rbac_rule_validation` decorator can only be applied to '
                    'an instance of `tempest.test.BaseTestCase`.')

            allowed = True
            disallowed_rules = []
            for rule in rules:
                _allowed = _is_authorized(
                    test_obj, service, rule, extra_target_data)
                if not _allowed:
                    disallowed_rules.append(rule)
                allowed = allowed and _allowed

            if disallowed_rules:
                # Choose the first disallowed rule and expect the error
                # code corresponding to it.
                first_error_index = rules.index(disallowed_rules[0])
                exp_error_code = expected_error_codes[first_error_index]
                LOG.debug("%s: Expecting %d to be raised for policy name: %s",
                          test_func.__name__, exp_error_code,
                          disallowed_rules[0])
            else:
                exp_error_code = expected_error_codes[0]

            expected_exception, irregular_msg = _get_exception_type(
                exp_error_code)

            caught_exception = None
            test_status = 'Allowed'

            try:
                test_func(*args, **kwargs)
            except rbac_exceptions.RbacInvalidServiceException:
                with excutils.save_and_reraise_exception():
                    msg = ("%s is not a valid service." % service)
                    # FIXME(felipemonteiro): This test_status is logged too
                    # late. Need a function to log it before re-raising.
                    test_status = ('Error, %s' % (msg))
                    LOG.error(msg)
            except (expected_exception,
                    rbac_exceptions.BasePatroleResponseBodyException) \
                    as actual_exception:
                caught_exception = actual_exception
                test_status = 'Denied'

                if irregular_msg:
                    LOG.warning(irregular_msg,
                                test_func.__name__,
                                ', '.join(rules),
                                service)

                if allowed:
                    msg = ("User with roles %s was not allowed to perform the "
                           "following actions: %s. Expected allowed actions: "
                           "%s. Expected disallowed actions: %s." % (
                               roles, sorted(rules),
                               sorted(set(rules) - set(disallowed_rules)),
                               sorted(disallowed_rules)))
                    LOG.error(msg)
                    raise rbac_exceptions.RbacUnderPermissionException(
                        "%s Exception was: %s" % (msg, actual_exception))
            except Exception as actual_exception:
                caught_exception = actual_exception

                if _check_for_expected_mismatch_exception(expected_exception,
                                                          actual_exception):
                    LOG.error('Expected and actual exceptions do not match. '
                              'Expected: %s. Actual: %s.',
                              expected_exception,
                              actual_exception.__class__)
                    raise rbac_exceptions.RbacExpectedWrongException(
                        expected=expected_exception,
                        actual=actual_exception.__class__,
                        exception=actual_exception)
                else:
                    with excutils.save_and_reraise_exception():
                        exc_info = sys.exc_info()
                        error_details = six.text_type(exc_info[1])
                        msg = ("An unexpected exception has occurred during "
                               "test: %s. Exception was: %s" % (
                                   test_func.__name__, error_details))
                        test_status = 'Error, %s' % (error_details)
                        LOG.error(msg)
            else:
                if not allowed:
                    msg = (
                        "OverPermission: Role %s was allowed to perform the "
                        "following disallowed actions: %s" % (
                            roles, sorted(disallowed_rules)
                        )
                    )
                    LOG.error(msg)
                    raise rbac_exceptions.RbacOverPermissionException(msg)
            finally:
                if CONF.patrole_log.enable_reporting:
                    RBACLOG.info(
                        "[Service]: %s, [Test]: %s, [Rules]: %s, "
                        "[Expected]: %s, [Actual]: %s",
                        service, test_func.__name__, ', '.join(rules),
                        "Allowed" if allowed else "Denied",
                        test_status)

                # Sanity-check that ``override_role`` was called to eliminate
                # false-positives and bad test flows resulting from exceptions
                # getting raised too early, too late or not at all, within
                # the scope of an RBAC test.
                _validate_override_role_called(
                    test_obj,
                    actual_exception=caught_exception)

        return wrapper
    return decorator


def _prepare_multi_policy(rules, exp_error_codes):
    if exp_error_codes:
        if not rules:
            msg = ("The `rules` list must be provided if using the "
                   "`expected_error_codes` list.")
            raise ValueError(msg)
        if len(rules) != len(exp_error_codes):
            msg = ("The `expected_error_codes` list is not the same length "
                   "as the `rules` list.")
            raise ValueError(msg)
        if not isinstance(exp_error_codes, (tuple, list)):
            exp_error_codes = [exp_error_codes]
    else:
        exp_error_codes = []

    if rules is None:
        rules = []
    elif not isinstance(rules, (tuple, list)):
        rules = [rules]

    # Fill in the exp_error_codes if needed. This is needed for the scenarios
    # where no exp_error_codes array is provided, so the error codes must be
    # set to the default error code value and there must be the same number
    # of error codes as rules.
    num_ecs = len(exp_error_codes)
    num_rules = len(rules)
    if (num_ecs < num_rules):
        for i in range(num_rules - num_ecs):
            exp_error_codes.append(_DEFAULT_ERROR_CODE)

    evaluated_rules = [
        r() if callable(r) else r for r in rules
    ]

    return evaluated_rules, exp_error_codes


def _is_authorized(test_obj, service, rule, extra_target_data):
    """Validates whether current RBAC role has permission to do policy action.

    :param test_obj: An instance or subclass of ``tempest.test.BaseTestCase``.
    :param service: The OpenStack service that enforces ``rule``.
    :param rule: The name of the policy action. Examples include
        "identity:create_user" or "os_compute_api:os-agents".
    :param extra_target_data: Dictionary, keyed with ``oslo.policy`` generic
        check names, whose values are string literals that reference nested
        ``tempest.test.BaseTestCase`` attributes. Used by ``oslo.policy`` for
        performing matching against attributes that are sent along with the API
        calls.

    :returns: True if the current RBAC role can perform the policy action,
        else False.

    :raises RbacResourceSetupFailed: If `project_id` or `user_id` are missing
        from the `auth_provider` attribute in `test_obj`.
    """

    try:
        project_id = test_obj.os_primary.credentials.project_id
        user_id = test_obj.os_primary.credentials.user_id
    except AttributeError as e:
        msg = ("{0}: project_id or user_id not found in os_primary.credentials"
               .format(e))
        LOG.error(msg)
        raise rbac_exceptions.RbacResourceSetupFailed(msg)

    roles = CONF.patrole.rbac_test_roles
    # TODO(vegasq) drop once CONF.patrole.rbac_test_role is removed
    if CONF.patrole.rbac_test_role:
        if not roles:
            roles.append(CONF.patrole.rbac_test_role)

    # Adding implied roles
    roles = test_obj.get_all_needed_roles(roles)

    # Test RBAC against custom requirements. Otherwise use oslo.policy.
    if CONF.patrole.test_custom_requirements:
        authority = requirements_authority.RequirementsAuthority(
            CONF.patrole.custom_requirements_file, service)
    else:
        formatted_target_data = _format_extra_target_data(
            test_obj, extra_target_data)
        authority = policy_authority.PolicyAuthority(
            project_id, user_id, service,
            extra_target_data=formatted_target_data)
    is_allowed = authority.allowed(rule, roles)

    if is_allowed:
        LOG.debug("[Policy action]: %s, [Role]: %s is allowed!", rule,
                  roles)
    else:
        LOG.debug("[Policy action]: %s, [Role]: %s is NOT allowed!",
                  rule, roles)

    return is_allowed


def _get_exception_type(expected_error_code=_DEFAULT_ERROR_CODE):
    """Dynamically calculate the expected exception to be caught.

    Dynamically calculate the expected exception to be caught by the test case.
    Only ``Forbidden`` and ``NotFound`` exceptions are permitted. ``NotFound``
    is supported because Neutron, for security reasons, masks ``Forbidden``
    exceptions as ``NotFound`` exceptions.

    :param expected_error_code: the integer representation of the expected
        exception to be caught. Must be contained in
        ``_SUPPORTED_ERROR_CODES``.
    :returns: tuple of the exception type corresponding to
        ``expected_error_code`` and a message explaining that a non-Forbidden
        exception was expected, if applicable.
    """
    expected_exception = None
    irregular_msg = None

    if not isinstance(expected_error_code, six.integer_types) \
            or expected_error_code not in _SUPPORTED_ERROR_CODES:
        msg = ("Please pass an expected error code. Currently "
               "supported codes: {0}".format(_SUPPORTED_ERROR_CODES))
        LOG.error(msg)
        raise rbac_exceptions.RbacInvalidErrorCode(msg)

    if expected_error_code == 403:
        expected_exception = lib_exc.Forbidden
    elif expected_error_code == 404:
        expected_exception = lib_exc.NotFound
        irregular_msg = ("NotFound exception was caught for test %s. Expected "
                         "policies which may have caused the error: %s. The "
                         "service %s throws a 404 instead of a 403, which is "
                         "irregular")
    return expected_exception, irregular_msg


def _format_extra_target_data(test_obj, extra_target_data):
    """Formats the "extra_target_data" dictionary with correct test data.

    Before being formatted, "extra_target_data" is a dictionary that maps a
    policy string like "trust.trustor_user_id" to a nested list of
    ``tempest.test.BaseTestCase`` attributes. For example, the attribute list
    in::

      "trust.trustor_user_id": "os.auth_provider.credentials.user_id"

    is parsed by iteratively calling ``getattr`` until the value of "user_id"
    is resolved. The resulting dictionary returns::

      "trust.trustor_user_id": "the user_id of the `os_primary` credential"

    :param test_obj: An instance or subclass of ``tempest.test.BaseTestCase``.
    :param extra_target_data: Dictionary, keyed with ``oslo.policy`` generic
        check names, whose values are string literals that reference nested
        ``tempest.test.BaseTestCase`` attributes. Used by ``oslo.policy`` for
        performing matching against attributes that are sent along with the API
        calls.
    :returns: Dictionary containing additional object data needed by
        ``oslo.policy`` to validate generic checks.
    """
    attr_value = test_obj
    formatted_target_data = {}

    for user_attribute, attr_string in extra_target_data.items():
        attrs = attr_string.split('.')
        for attr in attrs:
            attr_value = getattr(attr_value, attr)
        formatted_target_data[user_attribute] = attr_value

    return formatted_target_data


def _check_for_expected_mismatch_exception(expected_exception,
                                           actual_exception):
    """Checks that ``expected_exception`` matches ``actual_exception``.

    Since Patrole must handle 403/404 it is important that the expected and
    actual error codes match.

    :param excepted_exception: Expected exception for test.
    :param actual_exception: Actual exception raised by test.
    :returns: True if match, else False.
    :rtype: boolean
    """
    permission_exceptions = (lib_exc.Forbidden, lib_exc.NotFound)
    if isinstance(actual_exception, permission_exceptions):
        if not isinstance(actual_exception, expected_exception.__class__):
            return True
    return False


def _validate_override_role_called(test_obj, actual_exception):
    """Validates that :func:`rbac_utils.RbacUtilsMixin.override_role` is called
    during each Patrole test.

    Useful for validating that the expected exception isn't raised too early
    (before ``override_role`` call) or too late (after ``override_call``) or
    at all (which is a bad test).

    :param test_obj: An instance or subclass of ``tempest.test.BaseTestCase``.
    :param actual_exception: Actual exception raised by test.
    :raises RbacOverrideRoleException: If ``override_role`` isn't called, is
        called too early, or is called too late.
    """
    called = test_obj._validate_override_role_called()
    base_msg = ('This error is unrelated to RBAC and is due to either '
                'an API or override role failure. Exception: %s' %
                actual_exception)

    if not called:
        if actual_exception is not None:
            msg = ('Caught exception (%s) but it was raised before the '
                   '`override_role` context. ' % actual_exception.__class__)
        else:
            msg = 'Test missing required `override_role` call. '
        msg += base_msg
        LOG.error(msg)
        raise rbac_exceptions.RbacOverrideRoleException(msg)
    else:
        exc_caught_in_ctx = test_obj._validate_override_role_caught_exc()
        # This block is only executed if ``override_role`` is called. If
        # an exception is raised and the exception wasn't raised in the
        # ``override_role`` context and if the exception isn't a valid
        # exception type (instance of ``BasePatroleException``), then this is
        # a legitimate error.
        if (not exc_caught_in_ctx and
            actual_exception is not None and
            not isinstance(actual_exception,
                           rbac_exceptions.BasePatroleException)):
            msg = ('Caught exception (%s) but it was raised after the '
                   '`override_role` context. ' % actual_exception.__class__)
            msg += base_msg
            LOG.error(msg)
            raise rbac_exceptions.RbacOverrideRoleException(msg)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fixtures for Patrole tests."""
from __future__ import absolute_import

import fixtures
import mock
import time

from tempest.common import credentials_factory as credentials
from tempest import config
from tempest import test

from patrole_tempest_plugin import rbac_utils


CONF = config.CONF


class ConfPatcher(fixtures.Fixture):
    """Fixture to patch and restore global CONF. Adopted from Nova.

    This also resets overrides for everything that is patched during
    its teardown.
    """

    def __init__(self, **kwargs):
        """Constructor

        :params group: if specified all config options apply to that group.
        :params **kwargs: the rest of the kwargs are processed as a
            set of key/value pairs to be set as configuration override.
        """
        super(ConfPatcher, self).__init__()
        self.group = kwargs.pop('group', None)
        self.args = kwargs

    def setUp(self):
        super(ConfPatcher, self).setUp()
        for k, v in self.args.items():
            self.addCleanup(CONF.clear_override, k, self.group)
            CONF.set_override(k, v, self.group)


class FakeBaseRbacTest(rbac_utils.RbacUtilsMixin, test.BaseTestCase):
    os_primary = None

    def runTest(self):
        pass


class RbacUtilsMixinFixture(fixtures.Fixture):
    """Fixture for `RbacUtils` class."""

    USER_ID = mock.sentinel.user_id
    PROJECT_ID = mock.sentinel.project_id

    def __init__(self, do_reset_mocks=True, rbac_test_roles=None):
        self._do_reset_mocks = do_reset_mocks
        self._rbac_test_roles = rbac_test_roles or ['member']

    def patchobject(self, target, attribute, *args, **kwargs):
        p = mock.patch.object(target, attribute, *args, **kwargs)
        m = p.start()
        self.addCleanup(p.stop)
        return m

    def setUp(self):
        super(RbacUtilsMixinFixture, self).setUp()

        self.useFixture(ConfPatcher(rbac_test_roles=self._rbac_test_roles,
                                    group='patrole'))
        self.useFixture(ConfPatcher(
            admin_role='admin', auth_version='v3', group='identity'))
        self.useFixture(ConfPatcher(
            api_v3=True, group='identity-feature-enabled'))

        # Mock out functionality that can't be used by unit tests. Mocking out
        # time.sleep is a test optimization.
        self.mock_time = self.patchobject(rbac_utils, 'time',
                                          __name__='mock_time', spec=time)
        self.patchobject(credentials, 'get_configured_admin_credentials',
                         spec=object)
        mock_admin_mgr = self.patchobject(rbac_utils.clients, 'Manager',
                                          spec=rbac_utils.clients.Manager,
                                          roles_v3_client=mock.Mock(),
                                          roles_client=mock.Mock())
        self.admin_roles_client = mock_admin_mgr.return_value.roles_v3_client
        self.admin_roles_client.list_all_role_inference_rules.return_value = {
            "role_inferences": [
                {
                    "implies": [{"id": "reader_id", "name": "reader"}],
                    "prior_role": {"id": "member_id", "name": "member"}
                },
                {
                    "implies": [{"id": "member_id", "name": "member"}],
                    "prior_role": {"id": "admin_id", "name": "admin"}
                }
            ]
        }

        default_roles = {'admin', 'member', 'reader'}.union(
            set(self._rbac_test_roles))
        self.set_roles(list(default_roles), [])

        test_obj_kwargs = {
            'credentials.user_id': self.USER_ID,
            'credentials.tenant_id': self.PROJECT_ID,
            'credentials.project_id': self.PROJECT_ID,
        }

        class FakeRbacTest(FakeBaseRbacTest):
            os_primary = mock.Mock()

        FakeRbacTest.os_primary.configure_mock(**test_obj_kwargs)

        FakeRbacTest.setUpClass()
        self.test_obj = FakeRbacTest()
        if self._do_reset_mocks:
            self.admin_roles_client.reset_mock()
            self.test_obj.os_primary.reset_mock()
            self.mock_time.reset_mock()

    def set_roles(self, roles, roles_on_project=None):
        """Set the list of available roles in the system.

        :param roles: List of roles returned by ``list_roles``.
        :param roles_on_project: List of roles returned by
            ``list_user_roles_on_project``.
        :returns: None.
        """
        if not roles_on_project:
            roles_on_project = []
        if not isinstance(roles, list):
            roles = [roles]
        if not isinstance(roles_on_project, list):
            roles_on_project = [roles_on_project]

        available_roles = {
            'roles': [{'name': role, 'id': '%s_id' % role} for role in roles]
        }
        available_project_roles = {
            'roles': [{'name': role, 'id': '%s_id' % role}
                      for role in roles_on_project]
        }

        self.admin_roles_client.list_roles.return_value = available_roles
        self.admin_roles_client.list_user_roles_on_project.return_value = (
            available_project_roles)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.network import rbac_base as base


class SecGroupRbacTest(base.BaseNetworkRbacTest):

    @classmethod
    def skip_checks(cls):
        super(SecGroupRbacTest, cls).skip_checks()
        if not utils.is_extension_enabled('security-group', 'network'):
            msg = "security-group extension not enabled."
            raise cls.skipException(msg)

    @classmethod
    def resource_setup(cls):
        super(SecGroupRbacTest, cls).resource_setup()
        secgroup_name = data_utils.rand_name(cls.__name__ + '-secgroup')
        cls.secgroup = cls.security_groups_client.create_security_group(
            name=secgroup_name)['security_group']
        cls.addClassResourceCleanup(
            cls.security_groups_client.delete_security_group,
            cls.secgroup['id'])

    def _create_security_group(self):
        # Create a security group
        name = data_utils.rand_name(self.__class__.__name__ + '-secgroup')
        security_group =\
            self.security_groups_client.create_security_group(
                name=name)['security_group']
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            self.security_groups_client.delete_security_group,
            security_group['id'])
        return security_group

    def _create_security_group_rule(self):
        # Create a security group rule
        sec_group_rule = \
            self.security_group_rules_client.create_security_group_rule(
                security_group_id=self.secgroup['id'],
                direction='ingress',
                protocol='tcp',
                port_range_min=99,
                port_range_max=99)['security_group_rule']
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            self.security_group_rules_client.delete_security_group_rule,
            sec_group_rule['id'])
        return sec_group_rule

    @rbac_rule_validation.action(service="neutron",
                                 rules=["create_security_group"])
    @decorators.idempotent_id('db7003ce-5717-4e5b-afc7-befa35e8c67f')
    def test_create_security_group(self):

        with self.override_role():
            self._create_security_group()

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group"],
                                 expected_error_codes=[404])
    @decorators.idempotent_id('56335e77-aef2-4b54-86c7-7f772034b585')
    def test_show_security_group(self):

        with self.override_role():
            self.security_groups_client.show_security_group(
                self.secgroup['id'])

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group",
                                        "delete_security_group"],
                                 expected_error_codes=[404, 403])
    @decorators.idempotent_id('0b1330fd-dd28-40f3-ad73-966052e4b3de')
    def test_delete_security_group(self):

        # Create a security group
        secgroup_id = self._create_security_group()['id']

        with self.override_role():
            self.security_groups_client.delete_security_group(secgroup_id)

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group",
                                        "update_security_group"],
                                 expected_error_codes=[404, 403])
    @decorators.idempotent_id('56c5e4dc-f8aa-11e6-bc64-92361f002671')
    def test_update_security_group(self):

        # Create a security group
        secgroup_id = self._create_security_group()['id']

        with self.override_role():
            self.security_groups_client.update_security_group(
                secgroup_id,
                description="test description")

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group"])
    @decorators.idempotent_id('fbaf8d96-ed3e-49af-b24c-5fb44f05bbb7')
    def test_list_security_groups(self):
        """List Security Groups

        RBAC test for the neutron ``list_security_groups`` function and
        the ``get_security_group`` policy
        """
        admin_resource_id = self.secgroup['id']
        with (self.override_role_and_validate_list(
                admin_resource_id=admin_resource_id)) as ctx:
            ctx.resources = self.security_groups_client.list_security_groups(
                id=admin_resource_id)["security_groups"]

    @rbac_rule_validation.action(service="neutron",
                                 rules=["create_security_group_rule"])
    @decorators.idempotent_id('953d78df-00cd-416f-9cbd-b7cb4ea65772')
    def test_create_security_group_rule(self):

        with self.override_role():
            self._create_security_group_rule()

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group_rule",
                                        "delete_security_group_rule"],
                                 expected_error_codes=[404, 403])
    @decorators.idempotent_id('2262539e-b7d9-438c-acf9-a5ce0613be28')
    def test_delete_security_group_rule(self):

        sec_group_rule = self._create_security_group_rule()
        with self.override_role():
            self.security_group_rules_client.delete_security_group_rule(
                sec_group_rule['id'])

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group_rule"],
                                 expected_error_codes=[404])
    @decorators.idempotent_id('84b4038c-261e-4a94-90d5-c885739ab0d5')
    def test_show_security_group_rule(self):

        sec_group_rule = self._create_security_group_rule()
        with self.override_role():
            self.security_group_rules_client.show_security_group_rule(
                sec_group_rule['id'])

    @rbac_rule_validation.action(service="neutron",
                                 rules=["get_security_group_rule"])
    @decorators.idempotent_id('05739ab6-fa35-11e6-bc64-92361f002671')
    def test_list_security_group_rules(self):

        with self.override_role():
            security_rules = self.security_group_rules_client.\
                list_security_group_rules()

        # Neutron may return an empty list if access is denied.
        if not security_rules['security_group_rules']:
            raise rbac_exceptions.RbacEmptyResponseBody()



# Copyright 2017 AT&T Corporation
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base


class LimitsRbacTest(rbac_base.BaseV2ComputeRbacTest):

    @classmethod
    def skip_checks(cls):
        super(LimitsRbacTest, cls).skip_checks()
        if not utils.is_extension_enabled('os-limits', 'compute'):
            msg = "%s skipped as os-limits not enabled." % cls.__name__
            raise cls.skipException(msg)

    @rbac_rule_validation.action(service="nova",
                                 rules=["os_compute_api:limits"])
    @decorators.idempotent_id('3fb60f83-9a5f-4fdd-89d9-26c3710844a1')
    def test_show_limits(self):
        with self.override_role():
            self.limits_client.show_limits()



#    Copyright 2017 AT&T Corporation.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base


class ServerMetadataRbacTest(rbac_base.BaseV2ComputeRbacTest):

    @classmethod
    def setup_credentials(cls):
        cls.set_network_resources(network=True, subnet=True, router=True)
        super(ServerMetadataRbacTest, cls).setup_credentials()

    @classmethod
    def resource_setup(cls):
        super(ServerMetadataRbacTest, cls).resource_setup()
        cls.server = cls.create_test_server(metadata={}, wait_until='ACTIVE')
        cls.meta = {'default_key': 'value1', 'delete_key': 'value2'}

    def setUp(self):
        super(ServerMetadataRbacTest, self).setUp()
        self.servers_client.set_server_metadata(self.server['id'], self.meta)[
            'metadata']

    @decorators.idempotent_id('b07bbc27-58e2-4581-869d-ad228cec5d9a')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:index"])
    def test_list_server_metadata(self):
        with self.override_role():
            self.servers_client.list_server_metadata(self.server['id'])

    @decorators.idempotent_id('6e76748b-2417-4fa2-b41a-c0cc4bff356b')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:update_all"])
    def test_set_server_metadata(self):
        with self.override_role():
            self.servers_client.set_server_metadata(self.server['id'], {})

    @decorators.idempotent_id('1060bac4-fe16-4a77-be64-d8e482a06eab')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:create"])
    def test_update_server_metadata(self):
        with self.override_role():
            self.servers_client.update_server_metadata(self.server['id'], {})

    @decorators.idempotent_id('93dd8323-d3fa-48d1-8bd6-91c1b62fc341')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:show"])
    def test_show_server_metadata_item(self):
        with self.override_role():
            self.servers_client.show_server_metadata_item(
                self.server['id'], 'default_key')

    @decorators.idempotent_id('79511293-4bd7-447d-ba7e-634d0f4da70c')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:update"])
    def test_set_server_metadata_item(self):
        with self.override_role():
            self.servers_client.set_server_metadata_item(
                self.server['id'], 'default_key', {'default_key': 'value2'})

    @decorators.idempotent_id('feec5064-678d-40bc-a88f-c856e18d1e31')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:server-metadata:delete"])
    def test_delete_server_metadata_item(self):
        with self.override_role():
            self.servers_client.delete_server_metadata_item(
                self.server['id'], 'delete_key')



# Copyright 2017 AT&T Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from tempest.lib import exceptions

from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import requirements_authority as req_auth
from patrole_tempest_plugin.tests.unit import base


class BaseRequirementsAuthorityTest(base.TestCase):
    def setUp(self):
        super(BaseRequirementsAuthorityTest, self).setUp()
        self.rbac_auth = req_auth.RequirementsAuthority()
        self.current_directory = os.path.dirname(os.path.realpath(__file__))
        self.yaml_test_file = os.path.join(self.current_directory,
                                           'resources',
                                           'rbac_roles.yaml')
        self.expected_result = {'test:create': [['test_member'], ['_member_']],
                                'test:create2': [['test_member']],
                                'test:create3': [['test_member', '_member_']],
                                'test:create4': [['test_member', '!_member_']]}
        self.expected_rbac_map = {'test:create': ['test_member', '_member_'],
                                  'test:create2': ['test_member'],
                                  'test:create3': ['test_member, _member_'],
                                  'test:create4': ['test_member, !_member_']}


class RequirementsAuthorityTest(BaseRequirementsAuthorityTest):

    def test_requirements_auth_init(self):
        rbac_auth = req_auth.RequirementsAuthority(self.yaml_test_file, 'Test')
        self.assertEqual(self.expected_result, rbac_auth.roles_dict)

    def test_auth_allowed_empty_roles(self):
        self.rbac_auth.roles_dict = None
        self.assertRaises(exceptions.InvalidConfiguration,
                          self.rbac_auth.allowed, "", [""])

    def test_auth_allowed_role_in_api(self):
        self.rbac_auth.roles_dict = {'rule': [['_member_']]}
        self.assertTrue(self.rbac_auth.allowed("rule", ["_member_"]))

    def test_auth_allowed_role_not_in_api(self):
        self.rbac_auth.roles_dict = {'rule': [['_member_']]}
        self.assertFalse(self.rbac_auth.allowed("rule", "support_member"))

    def test_parser_get_allowed_invalid_rule_raises_parsing_exception(self):
        self.rbac_auth.roles_dict = {"foo": "bar"}
        self.assertRaises(rbac_exceptions.RbacParsingException,
                          self.rbac_auth.allowed, "baz", "support_member")

    def test_parser_init(self):
        req_auth.RequirementsParser(self.yaml_test_file)
        self.assertEqual([{'Test': self.expected_rbac_map}],
                         req_auth.RequirementsParser.Inner._rbac_map)

    def test_parser_role_in_api(self):
        req_auth.RequirementsParser.Inner._rbac_map = \
            [{'Test': self.expected_rbac_map}]
        self.rbac_auth.roles_dict = req_auth.RequirementsParser.parse("Test")

        self.assertEqual(self.expected_result, self.rbac_auth.roles_dict)
        self.assertTrue(
            self.rbac_auth.allowed("test:create2", ["test_member"]))

    def test_parser_role_not_in_api(self):
        req_auth.RequirementsParser.Inner._rbac_map = \
            [{'Test': self.expected_rbac_map}]
        self.rbac_auth.roles_dict = req_auth.RequirementsParser.parse("Test")

        self.assertEqual(self.expected_result, self.rbac_auth.roles_dict)
        self.assertFalse(self.rbac_auth.allowed("test:create2", "_member_"))

    def test_parser_except_invalid_configuration(self):
        req_auth.RequirementsParser.Inner._rbac_map = \
            [{'Test': self.expected_rbac_map}]
        self.rbac_auth.roles_dict = \
            req_auth.RequirementsParser.parse("Failure")

        self.assertFalse(self.rbac_auth.roles_dict)
        self.assertRaises(exceptions.InvalidConfiguration,
                          self.rbac_auth.allowed, "", [""])

    def test_auth_allowed_exclamation_mark_syntax_single_role(self):
        """Ensure that exclamation mark in front of role is dropped, and not
        considered as part of role itself.
        """

        self.rbac_auth.roles_dict = {'rule': [['!admin']]}
        self.assertTrue(self.rbac_auth.allowed("rule", ["member"]))
        self.assertTrue(self.rbac_auth.allowed("rule", ["!admin"]))
        self.assertFalse(self.rbac_auth.allowed("rule", ["admin"]))


class RequirementsAuthorityMultiRoleTest(BaseRequirementsAuthorityTest):

    def test_auth_allowed_exclamation_mark_syntax_multi_role(self):
        """Ensure that exclamation mark in front of role is dropped, and not
        considered as part of role itself.
        """

        self.rbac_auth.roles_dict = {'rule': [['member', '!admin']]}
        self.assertFalse(self.rbac_auth.allowed("rule", ["member", "admin"]))
        self.assertTrue(self.rbac_auth.allowed("rule", ["member", "!admin"]))

    def test_auth_allowed_single_rule_scenario(self):
        # member and support and not admin and not manager
        self.rbac_auth.roles_dict = {'rule': [['member', 'support',
                                               '!admin', '!manager']]}

        # User is member and support and not manager or admin
        self.assertTrue(self.rbac_auth.allowed("rule", ["member",
                                                        "support"]))

        # User is member and not manager or admin, but not support
        self.assertFalse(self.rbac_auth.allowed("rule", ["member"]))

        # User is support and not manager or admin, but not member
        self.assertFalse(self.rbac_auth.allowed("rule", ["support"]))

        # User is member and support and not manager, but have admin role
        self.assertFalse(self.rbac_auth.allowed("rule", ["member",
                                                         "support",
                                                         "admin"]))

        # User is member and not manager, but have admin role and not support
        self.assertFalse(self.rbac_auth.allowed("rule", ["member",
                                                         "admin"]))

        # User is member and support, but have manager and admin roles
        self.assertFalse(self.rbac_auth.allowed("rule", ["member",
                                                         "support",
                                                         "admin",
                                                         "manager"]))

    def test_auth_allowed_multi_rule_scenario(self):
        rules = [
            ['member', 'support', '!admin', '!manager'],
            ['member', 'admin'],
            ["manager"]
        ]
        self.rbac_auth.roles_dict = {'rule': rules}

        # Not a single role allows viewer
        self.assertFalse(self.rbac_auth.allowed("rule", ["viewer"]))
        # We have no rule that allows support and admin
        self.assertFalse(self.rbac_auth.allowed("rule", ["support",
                                                         "admin"]))
        # There is no rule that requires member without additional requirements
        self.assertFalse(self.rbac_auth.allowed("rule", ["member"]))

        # Pass with rules[2]
        self.assertTrue(self.rbac_auth.allowed("rule", ["manager"]))
        # Pass with rules[0]
        self.assertTrue(self.rbac_auth.allowed("rule", ["member",
                                                        "support"]))
        # Pass with rules[1]
        self.assertTrue(self.rbac_auth.allowed("rule", ["member",
                                                        "admin"]))
        # Pass with rules[2]
        self.assertTrue(self.rbac_auth.allowed("rule", ["manager",
                                                        "admin"]))
        # Pass with rules[1]
        self.assertTrue(self.rbac_auth.allowed("rule", ["member",
                                                        "support",
                                                        "admin"]))
        # Pass with rules[1]
        self.assertTrue(self.rbac_auth.allowed("rule", ["member",
                                                        "support",
                                                        "admin",
                                                        "manager"]))
        # Pass with rules[2]
        self.assertTrue(self.rbac_auth.allowed("rule", ["admin",
                                                        "manager"]))



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.image import rbac_base


class BasicOperationsImagesRbacTest(rbac_base.BaseV2ImageRbacTest):

    @classmethod
    def setup_clients(cls):
        super(BasicOperationsImagesRbacTest, cls).setup_clients()
        cls.image_client = cls.os_primary.image_client_v2

    def _create_image(self, **kwargs):
        image_name = data_utils.rand_name(self.__class__.__name__ + '-Image')
        image = self.create_image(name=image_name,
                                  container_format='bare',
                                  disk_format='raw',
                                  **kwargs)
        return image

    def _upload_image(self, image_id):
        image_file = six.BytesIO(data_utils.random_bytes())
        return self.image_client.store_image_file(image_id, image_file)

    @rbac_rule_validation.action(service="glance",
                                 rules=["add_image"])
    @decorators.idempotent_id('0f148510-63bf-11e6-b348-080027d0d606')
    def test_create_image(self):

        """Create Image Test

        RBAC test for the glance create_image endpoint
        """
        with self.override_role():
            self._create_image()

    @rbac_rule_validation.action(service="glance",
                                 rules=["upload_image"])
    @decorators.idempotent_id('fdc0c7e2-ad58-4c5a-ba9d-1f6046a5b656')
    def test_upload_image(self):

        """Upload Image Test

        RBAC test for the glance upload_image endpoint
        """
        image = self._create_image()

        with self.override_role():
            self._upload_image(image['id'])

    @decorators.idempotent_id('f0c268f3-cb51-49aa-9bd5-d30cf647322f')
    @rbac_rule_validation.action(service="glance",
                                 rules=["download_image"])
    def test_download_image(self):

        """Download Image Test

        RBAC test for the glance download_image endpoint
        """
        image = self._create_image()
        self._upload_image(image['id'])

        with self.override_role():
            self.image_client.show_image_file(image['id'])

    @rbac_rule_validation.action(service="glance",
                                 rules=["delete_image"])
    @decorators.idempotent_id('3b5c341e-645b-11e6-ac4f-080027d0d606')
    def test_delete_image(self):

        """Delete created image

        RBAC test for the glance delete_image endpoint
        """
        image = self._create_image()

        with self.override_role():
            self.image_client.delete_image(image['id'])
        self.image_client.wait_for_resource_deletion(image['id'])

    @rbac_rule_validation.action(service="glance",
                                 rules=["get_image"])
    @decorators.idempotent_id('3085c7c6-645b-11e6-ac4f-080027d0d606')
    def test_show_image(self):

        """Get created image

        RBAC test for the glance create_image endpoint
        """
        image = self._create_image()

        with self.override_role():
            self.image_client.show_image(image['id'])

    @rbac_rule_validation.action(service="glance",
                                 rules=["get_images"])
    @decorators.idempotent_id('bf1a4e94-645b-11e6-ac4f-080027d0d606')
    def test_list_images(self):

        """List all the images

        RBAC test for the glance list_images endpoint
        """
        with self.override_role():
            self.image_client.list_images()['images']

    @rbac_rule_validation.action(service="glance",
                                 rules=["modify_image"])
    @decorators.idempotent_id('32ecf48c-645e-11e6-ac4f-080027d0d606')
    def test_update_image(self):

        """Update given images

        RBAC test for the glance update_image endpoint
        """
        image = self._create_image()

        updated_image_name = data_utils.rand_name(
            self.__class__.__name__ + '-image')
        with self.override_role():
            self.image_client.update_image(image['id'], [
                dict(replace='/name', value=updated_image_name)])

    @decorators.idempotent_id('244050d9-1b9a-446a-b3c5-f26f3ba8eb75')
    @rbac_rule_validation.action(service="glance",
                                 rules=["modify_image"])
    def test_create_image_tag(self):

        """Create image tag

        RBAC test for the glance add_image_tag endpoint
        """
        image = self._create_image()

        with self.override_role():
            self.image_client.add_image_tag(
                image['id'],
                data_utils.rand_name(self.__class__.__name__ + '-tag'))

    @decorators.idempotent_id('c4a0bf9c-b78b-48c6-a31f-72c95f943c6e')
    @rbac_rule_validation.action(service="glance",
                                 rules=["modify_image"])
    def test_delete_image_tag(self):

        """Delete image tag

        RBAC test for the glance delete_image_tag endpoint
        """
        image = self._create_image()
        tag_name = data_utils.rand_name(self.__class__.__name__ + '-tag')
        self.image_client.add_image_tag(image['id'], tag_name)

        with self.override_role():
            self.image_client.delete_image_tag(image['id'], tag_name)

    @rbac_rule_validation.action(service="glance",
                                 rules=["publicize_image"])
    @decorators.idempotent_id('0ea4809c-6461-11e6-ac4f-080027d0d606')
    def test_publicize_image(self):

        """Publicize Image Test

        RBAC test for the glance publicize_image endpoint
        """
        with self.override_role():
            self._create_image(visibility='public')

    @decorators.idempotent_id('0f2d8427-134a-4d3c-a102-5fcdf5443d09')
    @rbac_rule_validation.action(service="glance",
                                 rules=["communitize_image"])
    def test_communitize_image(self):

        """Communitize Image Test

        RBAC test for the glance communitize_image policy
        """
        with self.override_role():
            self._create_image(visibility='community')

    @rbac_rule_validation.action(service="glance",
                                 rules=["deactivate"])
    @decorators.idempotent_id('b488458c-65df-11e6-9947-080027824017')
    def test_deactivate_image(self):

        """Deactivate Image Test

        RBAC test for the glance deactivate_image endpoint
        """
        image = self._create_image()
        self._upload_image(image['id'])

        with self.override_role():
            self.image_client.deactivate_image(image['id'])

    @rbac_rule_validation.action(service="glance",
                                 rules=["reactivate"])
    @decorators.idempotent_id('d3fa28b8-65df-11e6-9947-080027824017')
    def test_reactivate_image(self):

        """Reactivate Image Test

        RBAC test for the glance reactivate_image endpoint
        """
        image = self._create_image()
        self._upload_image(image['id'])

        with self.override_role():
            self.image_client.reactivate_image(image['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import waiters
from tempest import config
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.volume import rbac_base

CONF = config.CONF


class SnapshotsActionsV3RbacTest(rbac_base.BaseVolumeRbacTest):

    @classmethod
    def skip_checks(cls):
        super(SnapshotsActionsV3RbacTest, cls).skip_checks()
        if not CONF.volume_feature_enabled.snapshot:
            raise cls.skipException("Cinder snapshot feature disabled")

    @classmethod
    def resource_setup(cls):
        super(SnapshotsActionsV3RbacTest, cls).resource_setup()
        cls.volume = cls.create_volume()
        cls.snapshot = cls.create_snapshot(volume_id=cls.volume['id'])
        cls.snapshot_id = cls.snapshot['id']

    def tearDown(self):
        # Set snapshot's status to available after test
        status = 'available'
        self.snapshots_client.reset_snapshot_status(self.snapshot_id,
                                                    status)
        waiters.wait_for_volume_resource_status(self.snapshots_client,
                                                self.snapshot_id, status)
        super(SnapshotsActionsV3RbacTest, self).tearDown()

    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:snapshot_admin_actions:reset_status"])
    @decorators.idempotent_id('ea430145-34ef-408d-b678-95d5ae5f46eb')
    def test_reset_snapshot_status(self):
        status = 'error'
        with self.override_role():
            self.snapshots_client.reset_snapshot_status(
                self.snapshot['id'], status)
        waiters.wait_for_volume_resource_status(
            self.snapshots_client, self.snapshot['id'], status)

    @rbac_rule_validation.action(
        service="cinder",
        rules=["volume_extension:snapshot_admin_actions:force_delete"])
    @decorators.idempotent_id('a8b0f7d8-4c00-4645-b8d5-33ab4eecc6cb')
    def test_snapshot_force_delete(self):
        temp_snapshot = self.create_snapshot(self.volume['id'])

        with self.override_role():
            self.snapshots_client.force_delete_snapshot(temp_snapshot['id'])
        self.snapshots_client.wait_for_resource_deletion(temp_snapshot['id'])

    @decorators.idempotent_id('a95eab2a-c441-4609-9235-f7478627da88')
    @rbac_rule_validation.action(
        service="cinder",
        rules=["snapshot_extension:snapshot_actions:update_snapshot_status"])
    def test_update_snapshot_status(self):
        status = 'creating'
        self.snapshots_client.reset_snapshot_status(
            self.snapshot['id'], status)
        waiters.wait_for_volume_resource_status(self.snapshots_client,
                                                self.snapshot['id'], status)
        with self.override_role():
            self.snapshots_client.update_snapshot_status(self.snapshot['id'],
                                                         status="creating")
        waiters.wait_for_volume_resource_status(
            self.snapshots_client, self.snapshot['id'], status)



#    Copyright 2017 AT&T Corporation.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import identity
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base


class QuotaSetsRbacTest(rbac_base.BaseV2ComputeRbacTest):

    credentials = ['primary', 'admin']

    @classmethod
    def setup_clients(cls):
        super(QuotaSetsRbacTest, cls).setup_clients()
        cls.projects_client = cls.os_primary.projects_client

    @classmethod
    def skip_checks(cls):
        super(QuotaSetsRbacTest, cls).skip_checks()
        if not utils.is_extension_enabled('os-quota-sets', 'compute'):
            msg = "%s skipped as os-quota-sets extension not enabled."\
                  % cls.__name__
            raise cls.skipException(msg)

    @classmethod
    def resource_setup(cls):
        super(QuotaSetsRbacTest, cls).resource_setup()
        cls.tenant_id = cls.quotas_client.tenant_id
        cls.user_id = cls.quotas_client.user_id
        cls.quota_set = set(('injected_file_content_bytes',
                             'metadata_items', 'injected_files',
                             'ram', 'floating_ips', 'fixed_ips', 'key_pair',
                             'injected_file_path_bytes', 'instances',
                             'security_group_rules', 'cores',
                             'security_groups'))

    @decorators.idempotent_id('8229ceb0-db6a-4a2c-99c2-de226905d8b6')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-quota-sets:update"])
    def test_update_quota_set(self):
        default_quota_set = self.quotas_client.show_default_quota_set(
            self.tenant_id)['quota_set']
        default_quota_set.pop('id')
        new_quota_set = {'injected_file_content_bytes': 20480}

        with self.override_role():
            self.quotas_client.update_quota_set(self.tenant_id,
                                                force=True,
                                                **new_quota_set)
        self.addCleanup(self.quotas_client.update_quota_set, self.tenant_id,
                        **default_quota_set)

    @decorators.idempotent_id('58df5613-8f3c-400a-8b4b-2bae624d05e9')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-quota-sets:defaults"])
    def test_show_default_quota_set(self):
        with self.override_role():
            self.quotas_client.show_default_quota_set(self.tenant_id)

    @decorators.idempotent_id('e8169ac4-c402-4864-894e-aba74e3a459c')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-quota-sets:show"])
    def test_show_quota_set(self):
        with self.override_role():
            self.quotas_client.show_quota_set(self.tenant_id)

    @decorators.idempotent_id('4e240644-bf61-4872-9c32-8289ee2fdbbd')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-quota-sets:delete"])
    def test_delete_quota_set(self):
        project_name = data_utils.rand_name(
            self.__class__.__name__ + '-project')
        project_desc = project_name + '-desc'
        project = identity.identity_utils(self.os_admin).create_project(
            name=project_name, description=project_desc)
        project_id = project['id']
        self.addCleanup(
            identity.identity_utils(self.os_admin).delete_project,
            project_id)

        with self.override_role():
            self.quotas_client.delete_quota_set(project_id)

    @decorators.idempotent_id('ac9184b6-f3b3-4e17-a632-4b92c6500f86')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-quota-sets:detail"])
    def test_show_quota_set_details(self):
        with self.override_role():
            self.quotas_client.show_quota_set(self.tenant_id,
                                              detail=True)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import waiters
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base

from tempest import config

CONF = config.CONF


class VolumeRbacTest(rbac_base.BaseV2ComputeRbacTest):
    """RBAC tests for the Nova Volume client."""

    # These tests will fail with a 404 starting from microversion 2.36.
    # For more information, see:
    # https://developer.openstack.org/api-ref/compute/#volume-extension-os-volumes-os-snapshots-deprecated
    max_microversion = '2.35'

    @classmethod
    def skip_checks(cls):
        super(VolumeRbacTest, cls).skip_checks()
        if not CONF.service_available.cinder:
            skip_msg = ("%s skipped as Cinder is not available" % cls.__name__)
            raise cls.skipException(skip_msg)
        if not CONF.volume_feature_enabled.snapshot:
            skip_msg = ("Cinder volume snapshots are disabled")
            raise cls.skipException(skip_msg)

    @classmethod
    def resource_setup(cls):
        super(VolumeRbacTest, cls).resource_setup()
        cls.volume = cls.create_volume()

    def _delete_snapshot(self, snapshot_id):
        waiters.wait_for_volume_resource_status(
            self.snapshots_extensions_client, snapshot_id,
            'available')
        self.snapshots_extensions_client.delete_snapshot(snapshot_id)
        self.snapshots_extensions_client.wait_for_resource_deletion(
            snapshot_id)

    @decorators.idempotent_id('2402013e-a624-43e3-9518-44a5d1dbb32d')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_create_volume(self):
        with self.override_role():
            volume = self.volumes_extensions_client.create_volume(
                size=CONF.volume.volume_size)['volume']
        waiters.wait_for_volume_resource_status(self.volumes_client,
                                                volume['id'], 'available')
        # Use non-deprecated volumes_client for deletion.
        self.addCleanup(self.volumes_client.delete_volume, volume['id'])

    @decorators.idempotent_id('69b3888c-dff2-47b0-9fa4-0672619c9054')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_list_volumes(self):
        with self.override_role():
            self.volumes_extensions_client.list_volumes()

    @decorators.idempotent_id('4ba0a820-040f-488b-86bb-be2e920ea12c')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_show_volume(self):
        with self.override_role():
            self.volumes_extensions_client.show_volume(self.volume['id'])

    @decorators.idempotent_id('6e7870f2-1bb2-4b58-96f8-6782071ef327')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_delete_volume(self):
        volume = self.create_volume()
        with self.override_role():
            self.volumes_extensions_client.delete_volume(volume['id'])

    @decorators.idempotent_id('0c3eaa4f-69d6-4a13-9dda-19585f36b1c1')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_create_snapshot(self):
        s_name = data_utils.rand_name(self.__class__.__name__ + '-Snapshot')
        with self.override_role():
            snapshot = self.snapshots_extensions_client.create_snapshot(
                volume_id=self.volume['id'], display_name=s_name)['snapshot']
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self._delete_snapshot, snapshot['id'])

    @decorators.idempotent_id('e944e816-416c-11e7-a919-92ebcb67fe33')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_list_snapshots(self):
        with self.override_role():
            self.snapshots_extensions_client.list_snapshots()

    @decorators.idempotent_id('19c2e6bd-585b-472f-a8d7-71ea9299c655')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_show_snapshot(self):
        s_name = data_utils.rand_name(self.__class__.__name__ + '-Snapshot')
        snapshot = self.snapshots_extensions_client.create_snapshot(
            volume_id=self.volume['id'], display_name=s_name)['snapshot']
        self.addCleanup(self._delete_snapshot, snapshot['id'])

        with self.override_role():
            self.snapshots_extensions_client.show_snapshot(snapshot['id'])

    @decorators.idempotent_id('f4f5635c-416c-11e7-a919-92ebcb67fe33')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:os-volumes"])
    def test_delete_snapshot(self):
        s_name = data_utils.rand_name(self.__class__.__name__ + '-Snapshot')
        snapshot = self.snapshots_extensions_client.create_snapshot(
            volume_id=self.volume['id'], display_name=s_name)['snapshot']
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self._delete_snapshot, snapshot['id'])
        waiters.wait_for_volume_resource_status(
            self.snapshots_extensions_client, snapshot['id'],
            'available')

        with self.override_role():
            self.snapshots_extensions_client.delete_snapshot(snapshot['id'])
        self.snapshots_extensions_client.wait_for_resource_deletion(
            snapshot['id'])



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import waiters
from tempest.lib import decorators

from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.volume import rbac_base


class VolumesExtendV3RbacTest(rbac_base.BaseVolumeRbacTest):

    @classmethod
    def resource_setup(cls):
        super(VolumesExtendV3RbacTest, cls).resource_setup()
        # Create a test shared volume for tests
        cls.volume = cls.create_volume()

    @rbac_rule_validation.action(service="cinder", rules=["volume:extend"])
    @decorators.idempotent_id('1627b065-4081-4e14-8340-8e4fb02ceaf2')
    def test_volume_extend(self):
        # Extend volume test
        extend_size = int(self.volume['size']) + 1
        with self.override_role():
            self.volumes_client.extend_volume(self.volume['id'],
                                              new_size=extend_size)
        waiters.wait_for_volume_resource_status(
            self.volumes_client, self.volume['id'], 'available')



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import yaml

from oslo_log import log as logging

from tempest import config
from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin.rbac_authority import RbacAuthority
from patrole_tempest_plugin import rbac_exceptions

CONF = config.CONF
LOG = logging.getLogger(__name__)


class RequirementsParser(object):
    """A class that parses a custom requirements file."""
    _inner = None

    class Inner(object):
        _rbac_map = None

        def __init__(self, filepath):
            with open(filepath) as f:
                RequirementsParser.Inner._rbac_map = \
                    list(yaml.safe_load_all(f))

    def __init__(self, filepath):
        if RequirementsParser._inner is None:
            RequirementsParser._inner = RequirementsParser.Inner(filepath)

    @staticmethod
    def parse(component):
        """Parses a requirements file with the following format:

        .. code-block:: yaml

            <service_foo>:
              <api_action_a>:
                - <allowed_role_1>
                - <allowed_role_2>,<allowed_role_3>
                - <allowed_role_3>
              <api_action_b>:
                - <allowed_role_2>
                - <allowed_role_4>
            <service_bar>:
              <api_action_c>:
                - <allowed_role_3>

        :param str component: Name of the OpenStack service to be validated.
        :returns: The dictionary that maps each policy action to the list
            of allowed roles, for the given ``component``.
        :rtype: dict
        """
        try:
            for section in RequirementsParser.Inner._rbac_map:
                if component in section:
                    rules = copy.copy(section[component])

                    for rule in rules:
                        rules[rule] = [
                            roles.split(',') for roles in rules[rule]]

                        for i, role_pack in enumerate(rules[rule]):
                            rules[rule][i] = [r.strip() for r in role_pack]

                    return rules
        except yaml.parser.ParserError:
            LOG.error("Error while parsing the requirements YAML file. Did "
                      "you pass a valid component name from the test case?")
        return {}


class RequirementsAuthority(RbacAuthority):
    """A class that uses a custom requirements file to validate RBAC."""

    def __init__(self, filepath=None, component=None):
        """This class can be used to achieve a requirements-driven approach to
        validating an OpenStack cloud's RBAC implementation. Using this
        approach, Patrole computes expected test results by performing lookups
        against a custom requirements file which precisely defines the cloud's
        RBAC requirements.

        :param str filepath: Path where the custom requirements file lives.
            Defaults to ``[patrole].custom_requirements_file``.
        :param str component: Name of the OpenStack service to be validated.
        """
        self.filepath = filepath or CONF.patrole.custom_requirements_file
        if component is not None:
            self.roles_dict = RequirementsParser(self.filepath).parse(
                component)
        else:
            self.roles_dict = None

    def allowed(self, rule_name, roles):
        """Checks if a given rule in a policy is allowed with given role.

        :param string rule_name: Rule to be checked using provided requirements
            file specified by ``[patrole].custom_requirements_file``. Must be
            a key present in this file, under the appropriate component.
        :param List[string] roles: Roles to validate against custom
            requirements file.
        :returns: True if ``role`` is allowed to perform ``rule_name``, else
            False.
        :rtype: bool
        :raises RbacParsingException: If ``rule_name`` does not exist among the
            keyed policy names in the custom requirements file.
        """
        if not self.roles_dict:
            raise lib_exc.InvalidConfiguration(
                "Roles dictionary parsed from requirements YAML file is "
                "empty. Ensure the requirements YAML file is correctly "
                "formatted.")
        try:
            requirement_roles = self.roles_dict[rule_name]
        except KeyError:
            raise rbac_exceptions.RbacParsingException(
                "'%s' rule name is not defined in the requirements YAML file: "
                "%s" % (rule_name, self.filepath))

        for role_reqs in requirement_roles:
            required_roles = [
                role for role in role_reqs if not role.startswith("!")]
            forbidden_roles = [
                role[1:] for role in role_reqs if role.startswith("!")]

            # User must have all required roles
            required_passed = all([r in roles for r in required_roles])
            # User must not have any forbidden roles
            forbidden_passed = all([r not in forbidden_roles
                                    for r in roles])

            if required_passed and forbidden_passed:
                return True

        return False



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tempest.common import image as common_image
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from patrole_tempest_plugin import rbac_exceptions
from patrole_tempest_plugin import rbac_rule_validation
from patrole_tempest_plugin.tests.api.compute import rbac_base

CONF = config.CONF


class ImagesRbacTest(rbac_base.BaseV2ComputeRbacTest):
    """RBAC tests for the Nova images API.

    These APIs are proxy calls to the Image service. Consequently, no Nova
    policy actions are enforced; instead, only Glance policy actions are
    enforced. As such, these tests check that only Glance policy actions are
    executed.
    """

    # These tests will fail with a 404 starting from microversion 2.36.
    # See the following link for details:
    # https://developer.openstack.org/api-ref/compute/#images-deprecated
    max_microversion = '2.35'

    @classmethod
    def skip_checks(cls):
        super(ImagesRbacTest, cls).skip_checks()
        if not CONF.service_available.glance:
            skip_msg = ("%s skipped as glance is not available" % cls.__name__)
            raise cls.skipException(skip_msg)

    @classmethod
    def setup_clients(cls):
        super(ImagesRbacTest, cls).setup_clients()
        if CONF.image_feature_enabled.api_v2:
            cls.glance_image_client = cls.os_primary.image_client_v2
        elif CONF.image_feature_enabled.api_v1:
            cls.glance_image_client = cls.os_primary.image_client
        else:
            raise lib_exc.InvalidConfiguration(
                'Either api_v1 or api_v2 must be True in '
                '[image-feature-enabled].')

    @classmethod
    def resource_setup(cls):
        super(ImagesRbacTest, cls).resource_setup()
        params = {'name': data_utils.rand_name(cls.__name__ + '-image')}
        if CONF.image_feature_enabled.api_v1:
            params = {'headers': common_image.image_meta_to_headers(**params)}

        cls.image = cls.glance_image_client.create_image(**params)
        cls.addClassResourceCleanup(
            cls.glance_image_client.wait_for_resource_deletion,
            cls.image['id'])
        cls.addClassResourceCleanup(
            cls.glance_image_client.delete_image, cls.image['id'])

    @decorators.idempotent_id('b861f302-b72b-4055-81db-c62ff30b136d')
    @rbac_rule_validation.action(
        service="glance",
        rules=["get_images"])
    def test_list_images(self):
        with self.override_role():
            self.compute_images_client.list_images()

    @decorators.idempotent_id('4365ae0f-15ee-4b54-a527-1679faaed140')
    @rbac_rule_validation.action(
        service="glance",
        rules=["get_images"])
    def test_list_images_with_details(self):
        with self.override_role():
            self.compute_images_client.list_images(detail=True)

    @decorators.idempotent_id('886dfcae-51bf-4610-9e52-82d7189524c2')
    @rbac_rule_validation.action(
        service="glance",
        rules=["get_image"])
    def test_show_image_details(self):
        with self.override_role():
            self.compute_images_client.show_image(self.image['id'])

    @decorators.idempotent_id('5888c7aa-0803-46d4-a3fb-5d4729465cd5')
    @rbac_rule_validation.action(
        service="glance",
        rules=["delete_image"])
    def test_delete_image(self):
        image = self.glance_image_client.create_image(
            name=data_utils.rand_name(self.__class__.__name__ + '-image'))
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.glance_image_client.delete_image, image['id'])

        with self.override_role():
            self.compute_images_client.delete_image(image['id'])


class ImagesMetadataRbacTest(rbac_base.BaseV2ComputeRbacTest):
    """RBAC tests for the Nova metadata images API.

    These APIs are proxy calls to the Image service. Consequently, no Nova
    policy actions are enforced; instead, only Glance policy actions are
    enforced. As such, these tests check that only Glance policy actions are
    executed.
    """

    # These tests will fail with a 404 starting from microversion 2.39.
    # See the following link for details:
    # https://developer.openstack.org/api-ref/compute/#images-deprecated
    max_microversion = '2.38'

    @classmethod
    def skip_checks(cls):
        super(ImagesMetadataRbacTest, cls).skip_checks()
        if not CONF.service_available.glance:
            skip_msg = ("%s skipped as glance is not available" % cls.__name__)
            raise cls.skipException(skip_msg)

    @classmethod
    def setup_clients(cls):
        super(ImagesMetadataRbacTest, cls).setup_clients()
        if CONF.image_feature_enabled.api_v2:
            cls.glance_image_client = cls.os_primary.image_client_v2
        elif CONF.image_feature_enabled.api_v1:
            cls.glance_image_client = cls.os_primary.image_client
        else:
            raise lib_exc.InvalidConfiguration(
                'Either api_v1 or api_v2 must be True in '
                '[image-feature-enabled].')

    @classmethod
    def resource_setup(cls):
        super(ImagesMetadataRbacTest, cls).resource_setup()
        params = {'name': data_utils.rand_name(cls.__name__ + '-image')}
        if CONF.image_feature_enabled.api_v1:
            params = {'headers': common_image.image_meta_to_headers(**params)}

        cls.image = cls.glance_image_client.create_image(**params)
        cls.addClassResourceCleanup(
            cls.glance_image_client.wait_for_resource_deletion,
            cls.image['id'])
        cls.addClassResourceCleanup(
            cls.glance_image_client.delete_image, cls.image['id'])

    @decorators.idempotent_id('dbe09d4c-e615-48cb-b908-a06a0f410a8e')
    @rbac_rule_validation.action(
        service="glance",
        rules=["get_image"])
    def test_show_image_metadata_item(self):
        self.compute_images_client.set_image_metadata(self.image['id'],
                                                      meta={'foo': 'bar'})
        self.addCleanup(self.compute_images_client.delete_image_metadata_item,
                        self.image['id'], key='foo')

        with self.override_role():
            self.compute_images_client.show_image_metadata_item(
                self.image['id'], key='foo')

    @decorators.idempotent_id('59f66079-d564-47e8-81b0-03c2e84d339e')
    @rbac_rule_validation.action(
        service="glance",
        rules=["get_image"])
    def test_list_image_metadata(self):
        with self.override_role():
            self.compute_images_client.list_image_metadata(self.image['id'])

    @decorators.idempotent_id('575604aa-909f-4b1b-a5a5-cfae1f63044b')
    @rbac_rule_validation.action(
        service="glance",
        rules=["modify_image"])
    def test_create_image_metadata(self):
        with self.override_role():
            # NOTE(felipemonteiro): Although the name of the client function
            # appears wrong, it's actually correct: update_image_metadata does
            # an http post.
            self.compute_images_client.update_image_metadata(
                self.image['id'], meta={'foo': 'bar'})
        self.addCleanup(self.compute_images_client.delete_image_metadata_item,
                        self.image['id'], key='foo')

    @decorators.idempotent_id('fb8c4eb6-00e5-454c-b8bc-0e801ec369f1')
    @rbac_rule_validation.action(
        service="glance",
        rules=["modify_image"])
    def test_update_image_metadata(self):
        with self.override_role():
            self.compute_images_client.set_image_metadata(self.image['id'],
                                                          meta={'foo': 'bar'})
        self.addCleanup(self.compute_images_client.delete_image_metadata_item,
                        self.image['id'], key='foo')

    @decorators.idempotent_id('9c7c2036-af9b-49a8-8ba1-09b027ee5def')
    @rbac_rule_validation.action(
        service="glance",
        rules=["modify_image"])
    def test_update_image_metadata_item(self):
        with self.override_role():
            self.compute_images_client.set_image_metadata_item(
                self.image['id'], meta={'foo': 'bar'}, key='foo')
        self.addCleanup(self.compute_images_client.delete_image_metadata_item,
                        self.image['id'], key='foo')

    @decorators.idempotent_id('5f0dc4e6-0761-4613-9bde-0a6acdc78f46')
    @rbac_rule_validation.action(
        service="glance",
        rules=["modify_image"])
    def test_delete_image_metadata_item(self):
        self.compute_images_client.set_image_metadata(self.image['id'],
                                                      meta={'foo': 'bar'})
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.compute_images_client.delete_image_metadata_item,
                        self.image['id'], key='foo')

        with self.override_role():
            self.compute_images_client.delete_image_metadata_item(
                self.image['id'], key='foo')


class ImageSizeRbacTest(rbac_base.BaseV2ComputeRbacTest):
    """Tests the ``image_size`` compute policies.

    NOTE(felipemonteiro): If Patrole is enhanced to test multiple policies
    simultaneously, these policy actions can be combined with the related
    tests from ``ImagesRbacTest`` above.
    """

    # These tests will fail with a 404 starting from microversion 2.36.
    # See the following link for details:
    # https://developer.openstack.org/api-ref/compute/#images-deprecated
    max_microversion = '2.35'

    @classmethod
    def skip_checks(cls):
        super(ImageSizeRbacTest, cls).skip_checks()
        if not CONF.service_available.glance:
            skip_msg = ("%s skipped as glance is not available" % cls.__name__)
            raise cls.skipException(skip_msg)

    @classmethod
    def setup_clients(cls):
        super(ImageSizeRbacTest, cls).setup_clients()
        if CONF.image_feature_enabled.api_v2:
            cls.glance_image_client = cls.os_primary.image_client_v2
        elif CONF.image_feature_enabled.api_v1:
            cls.glance_image_client = cls.os_primary.image_client
        else:
            raise lib_exc.InvalidConfiguration(
                'Either api_v1 or api_v2 must be True in '
                '[image-feature-enabled].')

    @classmethod
    def resource_setup(cls):
        super(ImageSizeRbacTest, cls).resource_setup()
        params = {'name': data_utils.rand_name(cls.__name__ + '-image')}
        if CONF.image_feature_enabled.api_v1:
            params = {'headers': common_image.image_meta_to_headers(**params)}

        cls.image = cls.glance_image_client.create_image(**params)
        cls.addClassResourceCleanup(
            cls.glance_image_client.wait_for_resource_deletion,
            cls.image['id'])
        cls.addClassResourceCleanup(
            cls.glance_image_client.delete_image, cls.image['id'])

    @testtools.skipIf(CONF.policy_feature_enabled.removed_nova_policies_stein,
                      "This API extension policy was removed in Stein")
    @decorators.idempotent_id('fe34d2a6-5743-45bf-8f92-a1d703d7c7ab')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:image-size"])
    def test_show_image_includes_image_size(self):
        with self.override_role():
            body = self.compute_images_client.show_image(self.image['id'])[
                'image']

        expected_attr = 'OS-EXT-IMG-SIZE:size'
        if expected_attr not in body:
            raise rbac_exceptions.RbacMissingAttributeResponseBody(
                attribute=expected_attr)

    @testtools.skipIf(CONF.policy_feature_enabled.removed_nova_policies_stein,
                      "This API extension policy was removed in Stein")
    @decorators.idempotent_id('08342c7d-297d-42ee-b398-90fce2443792')
    @rbac_rule_validation.action(
        service="nova",
        rules=["os_compute_api:image-size"])
    def test_list_images_with_details_includes_image_size(self):
        with self.override_role():
            body = self.compute_images_client.list_images(detail=True)[
                'images']

        expected_attr = 'OS-EXT-IMG-SIZE:size'
        if expected_attr not in body[0]:
            raise rbac_exceptions.RbacMissingAttributeResponseBody(
                attribute=expected_attr)
